from ScoutSuite.providers.gcp.resources.projects import Projects
from ScoutSuite.providers.gcp.resources.gke.clusters import Clusters


class KubernetesEngine(Projects):
    _children = [
        (Clusters, 'clusters')
    ]

    async def fetch_all(self):
        await Projects.fetch_all(self)



from ScoutSuite.providers.base.resources.base import Resources
from ScoutSuite.providers.gcp.facade.base import GCPFacade


class MonitoringAlertPolicies(Resources):
    def __init__(self, facade: GCPFacade, project_id: str):
        super().__init__(facade)
        self.project_id = project_id

    async def fetch_all(self):
        raw_alert_policies = await self.facade.stackdrivermonitoring.get_alert_policies(self.project_id)
        alert_policy = self._parse_alert_policy(raw_alert_policies)
        self[self.project_id] = alert_policy

    def _parse_alert_policy(self, raw_alert_policies):
        alert_policy_dict = {}
        alert_policy_dict['project_ownership_assignments'] = \
            self._specific_alert_policy_present(raw_alert_policies)
        alert_policy_dict['audit_config_change'] = self._specific_alert_policy_present(raw_alert_policies)
        alert_policy_dict['custom_role_change'] = self._specific_alert_policy_present(raw_alert_policies)
        alert_policy_dict['vpc_network_firewall_rule_change'] = self._specific_alert_policy_present(raw_alert_policies)
        alert_policy_dict['vpc_network_route_change'] = self._specific_alert_policy_present(raw_alert_policies)
        alert_policy_dict['vpc_network_change'] = self._specific_alert_policy_present(raw_alert_policies)
        alert_policy_dict['cloud_storage_iam_permission_change'] = \
            self._specific_alert_policy_present(raw_alert_policies)
        alert_policy_dict['sql_instance_conf_change'] = self._specific_alert_policy_present(raw_alert_policies)
        return alert_policy_dict

    def _specific_alert_policy_present(self, alert_policies):
        for alert_policy in alert_policies:
            for condition in alert_policy.conditions:
                if condition.condition_threshold.filter == 'metric.type=\"logging.googleapis.com/user/<Log Metric ' \
                                                           'Name>\"' and alert_policy.enabled.value:
                    return True
        return False



from ScoutSuite.providers.azure.resources.subscriptions import Subscriptions

from .auto_provisioning_settings import AutoProvisioningSettings
from .pricings import Pricings
# from .alerts import Alerts
from .security_contacts import SecurityContacts
# from .information_protection_policies import InformationProtectionPolicies
# from .settings import Settings
from .compliance_results import ComplianceResults
from .regulatory_compliance_results import RegulatoryComplianceResults


class SecurityCenter(Subscriptions):
    _children = [
        (AutoProvisioningSettings, 'auto_provisioning_settings'),
        (Pricings, 'pricings'),
        # (Alerts, 'alerts'),  # FIXME this needs to be tested with alert results...
        (SecurityContacts, 'security_contacts'),
        # (InformationProtectionPolicies, 'information_protection_policies'),  # FIXME this isn't properly implemented
        # (Settings, 'settings')  # FIXME this isn't implemented
        (ComplianceResults, 'compliance_results'),
        (RegulatoryComplianceResults, 'regulatory_compliance_results')
    ]



from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureResources
from ScoutSuite.providers.utils import get_non_provider_id


class Snapshots(AzureResources):

    def __init__(self, facade: AzureFacade, subscription_id: str):
        super().__init__(facade)
        self.subscription_id = subscription_id

    async def fetch_all(self):
        for raw_snapshot in await self.facade.virtualmachines.get_snapshots(self.subscription_id):
            id, snapshot = self._parse_snapshot(raw_snapshot)
            self[id] = snapshot

    def _parse_snapshot(self, raw_snapshot):
        snapshot_dict = {}

        snapshot_dict['id'] = get_non_provider_id(raw_snapshot.id)
        snapshot_dict['unique_id'] = getattr(raw_snapshot, 'unique_id', None)
        snapshot_dict['name'] = raw_snapshot.name
        snapshot_dict['type'] = raw_snapshot.type
        snapshot_dict['location'] = raw_snapshot.location
        snapshot_dict['tags'] = raw_snapshot.tags
        snapshot_dict['managed_by'] = raw_snapshot.managed_by
        snapshot_dict['sku'] = raw_snapshot.sku
        snapshot_dict['time_created'] = raw_snapshot.time_created
        snapshot_dict['os_type'] = raw_snapshot.os_type
        snapshot_dict['hyper_vgeneration'] = raw_snapshot.hyper_vgeneration
        snapshot_dict['creation_data'] = raw_snapshot.creation_data
        snapshot_dict['disk_size_gb'] = raw_snapshot.disk_size_gb
        snapshot_dict['disk_size_bytes'] = getattr(raw_snapshot, 'disk_size_bytes', None)
        snapshot_dict['provisioning_state'] = raw_snapshot.provisioning_state
        snapshot_dict['incremental'] = getattr(raw_snapshot, 'incremental', None)
        snapshot_dict['additional_properties'] = raw_snapshot.additional_properties

        if hasattr(raw_snapshot, 'encryption'):
            snapshot_dict['encryption_type'] = getattr(raw_snapshot.encryption, 'type', None)
        else:
            snapshot_dict['encryption_type'] = None

        return snapshot_dict['id'], snapshot_dict




from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.base import AWSResources
from ScoutSuite.providers.utils import get_non_provider_id


class Distributions(AWSResources):
    async def fetch_all(self):
        list_distributions = await self.facade.cloudfront.get_distributions()
        for distribution in list_distributions:
            id, distro = self._parse_distributions(distribution)
            self[id] = distro

    def _parse_distributions(self, raw_distribution):
        distribution_dict = {}
        distribution_dict['id'] = distribution_dict['name'] = raw_distribution.get('Id')
        distribution_dict['arn'] = raw_distribution.get('ARN')
        distribution_dict['aliases'] = raw_distribution.get('Aliases')
        distribution_dict['status'] = raw_distribution.get('Status')
        distribution_dict['cache_behaviors'] = raw_distribution.get('CacheBehaviors')
        distribution_dict['restrictions'] = raw_distribution.get('Restrictions')
        distribution_dict['origins'] = raw_distribution.get('Origins')
        distribution_dict['domain_name'] = raw_distribution.get('DomainName')
        distribution_dict['web_acl_id'] = raw_distribution.get('WebACLId')
        distribution_dict['price_class'] = raw_distribution.get('PriceClass')
        distribution_dict['enabled'] = raw_distribution.get('Enabled')
        distribution_dict['default_cache_behavior'] = raw_distribution.get('DefaultCacheBehavior')
        distribution_dict['is_ipv6_enabled'] = raw_distribution.get('IsIPV6Enabled')
        distribution_dict['comment'] = raw_distribution.get('Comment')
        distribution_dict['http_version'] = raw_distribution.get('HttpVersion')
        distribution_dict['viewer_certificate'] = raw_distribution.get('ViewerCertificate')
        distribution_dict['custom_error_responses'] = raw_distribution.get('CustomErrorResponses')
        distribution_dict['last_modified_time'] = raw_distribution.get('LastModifiedTime')
        distribution_dict['origin_groups'] = raw_distribution.get('OriginGroups')
        return distribution_dict['id'], distribution_dict




from google.cloud import logging as stackdriverlogging
from google.api_core.gapic_v1.client_info import ClientInfo

from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.utils import run_concurrently
from ScoutSuite.utils import get_user_agent


class StackdriverLoggingFacade:

    def get_client(self, project_id: str):
        client_info = ClientInfo(user_agent=get_user_agent())
        client = stackdriverlogging.Client(project=project_id,
                                           client_info=client_info)
        return client

    async def get_sinks(self, project_id: str):
        try:
            client = self.get_client(project_id)
            return await run_concurrently(lambda: [sink for sink in client.list_sinks()])
        except Exception as e:
            print_exception(f'Failed to retrieve sinks: {e}')
            return []

    async def get_metrics(self, project_id: str):
        try:
            client = self.get_client(project_id)
            return await run_concurrently(lambda: [metric for metric in client.list_metrics()])
        except Exception as e:
            print_exception(f'Failed to retrieve metrics: {e}')
            return []



from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureCompositeResources
from ScoutSuite.providers.azure.utils import get_resource_group_name

from .database_blob_auditing_policies import DatabaseBlobAuditingPolicies
from .database_threat_detection_policies import DatabaseThreatDetectionPolicies
from .replication_links import ReplicationLinks
from .transparent_data_encryptions import TransparentDataEncryptions


class Databases(AzureCompositeResources):
    _children = [
        (DatabaseBlobAuditingPolicies, 'auditing'),
        (DatabaseThreatDetectionPolicies, 'threat_detection'),
        (ReplicationLinks, None),
        (TransparentDataEncryptions, None)
    ]

    def __init__(self, facade: AzureFacade, resource_group_name: str, server_name: str, subscription_id: str):
        super().__init__(facade)
        self.resource_group_name = resource_group_name
        self.server_name = server_name
        self.subscription_id = subscription_id

    async def fetch_all(self):
        for db in await self.facade.sqldatabase.get_databases(
                self.resource_group_name, self.server_name, self.subscription_id):
            # We do not want to scan 'master' database which is auto-generated by Azure and read-only:
            if db.name == 'master':
                continue

            self[db.name] = {
                'id': db.name,
                'name': db.name,
                'tags': ["{}:{}".format(key, value) for key, value in  db.tags.items()] if db.tags is not None else [],
                'resource_group_name': get_resource_group_name(db.id)
            }

        await self._fetch_children_of_all_resources(
            resources=self,
            scopes={db_id: {'resource_group_name': self.resource_group_name,
                            'server_name': self.server_name,
                            'database_name': db['name'],
                            'subscription_id': self.subscription_id}
                    for (db_id, db) in self.items()}
        )



from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.base.resources.base import Resources
from ScoutSuite.providers.gcp.facade.base import GCPFacade
from ScoutSuite.providers.utils import get_non_provider_id


class KMSPolicy(Resources):

    def __init__(self, facade: GCPFacade, project_id: str, keyring_name: str, location: str, key_name: str):
        super().__init__(facade)
        self.project_id = project_id
        self.keyring_name = keyring_name
        self.location = location
        self.key_name = key_name

    async def fetch_all(self):
        raw_kms_bindings = await self.facade.kms.keys_iam_policy(self.project_id, self.location, self.keyring_name, self.key_name)
        for raw_kms_binding in raw_kms_bindings:
            kms_binding_id, kms_bindings = await self._parse_binding(raw_kms_binding)
            self[kms_binding_id] = kms_bindings

    async def _parse_binding(self, kms_raw_binding):
        kms_binding_dict = {}
        kms_binding_dict['id'] = get_non_provider_id(kms_raw_binding['role'])
        kms_binding_dict['name'] = kms_raw_binding['role'].split('/')[-1]
        kms_binding_dict['members'] = kms_raw_binding['members']
        kms_binding_dict['custom_role'] = 'projects/' in kms_raw_binding['role']
        kms_binding_dict['anonymous_public_accessible'] = self.keys_not_anonymous_public_accessible(kms_raw_binding)

        role_definition = await self.facade.iam.get_role_definition(kms_raw_binding['role'])

        kms_binding_dict['title'] = role_definition.get('title')
        kms_binding_dict['description'] = role_definition.get('description')
        kms_binding_dict['permissions'] = role_definition.get('includedPermissions')

        return kms_binding_dict['id'], kms_binding_dict

    def keys_not_anonymous_public_accessible(self, kms_raw_binding):
        if 'allUsers' in kms_raw_binding['members'] or 'allAuthenticatedUsers' in kms_raw_binding['members']:
            return False
        return True



from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.gcp.facade.basefacade import GCPBaseFacade
from ScoutSuite.providers.gcp.facade.utils import GCPFacadeUtils


class MemoryStoreRedisFacade(GCPBaseFacade):
    def __init__(self):
        super().__init__('redis', 'v1beta1')

    async def get_redis_instances(self, project_id: str):
        # Retrieves Redis Instances using the Cloud Memorystore API
        try:
            formatted_parent = f'projects/{project_id}/locations/-'
            cloudmem_client = self._get_client()
            instances_group = cloudmem_client.projects().locations().instances()
            request = instances_group.list(parent=formatted_parent)
            return await GCPFacadeUtils.get_all('instances', request, instances_group)
        except Exception as e:
            print_exception(f'Failed to retrieve redis instances: {e}')
            return []



import json
import logging
import requests
from getpass import getpass
from datetime import datetime, timedelta

from azure.common.credentials import ServicePrincipalCredentials, UserPassCredentials, get_azure_cli_credentials
from msrestazure.azure_active_directory import MSIAuthentication
from ScoutSuite.core.console import print_info, print_debug, print_exception
from msrestazure.azure_active_directory import AADTokenCredentials
import adal

from ScoutSuite.providers.base.authentication_strategy import AuthenticationStrategy, AuthenticationException


AUTHORITY_HOST_URI = 'https://login.microsoftonline.com'
AZURE_CLI_CLIENT_ID = "04b07795-8ddb-461a-bbee-02f9e1bf7b46"


class AzureCredentials:

    def __init__(self,
                 arm_credentials, aad_graph_credentials,
                 tenant_id=None, default_subscription_id=None,
                 context=None):

        self.arm_credentials = arm_credentials  # Azure Resource Manager API credentials
        self.aad_graph_credentials = aad_graph_credentials  # Azure AD Graph API credentials
        self.tenant_id = tenant_id
        self.default_subscription_id = default_subscription_id
        self.context = context

    def get_tenant_id(self):
        if self.tenant_id:
            return self.tenant_id
        elif 'tenant_id' in self.aad_graph_credentials.token:
            return self.aad_graph_credentials.token['tenant_id']
        else:
            # This is a last resort, e.g. for MSI authentication
            try:
                h = {'Authorization': 'Bearer {}'.format(self.arm_credentials.token['access_token'])}
                r = requests.get('https://management.azure.com/tenants?api-version=2020-01-01', headers=h)
                r2 = r.json()
                return r2.get('value')[0].get('tenantId')
            except Exception as e:
                print_exception('Unable to infer tenant ID: {}'.format(e))
                return None

    def get_credentials(self, resource):
        if resource == 'arm':
            self.arm_credentials = self.get_fresh_credentials(self.arm_credentials)
            return self.arm_credentials
        elif resource == 'aad_graph':
            self.aad_graph_credentials = self.get_fresh_credentials(self.aad_graph_credentials)
            return self.aad_graph_credentials
        else:
            raise AuthenticationException('Invalid credentials resource type')

    def get_fresh_credentials(self, credentials):
        """
        Check if credentials are outdated and if so refresh them.
        """
        if self.context and hasattr(credentials, 'token'):
            expiration_datetime = datetime.fromtimestamp(credentials.token['expires_on'])
            current_datetime = datetime.now()
            expiration_delta = expiration_datetime - current_datetime
            if expiration_delta < timedelta(minutes=5):
                return self.refresh_credential(credentials)
        return credentials

    def refresh_credential(self, credentials):
        """
        Refresh credentials
        """
        print_debug('Refreshing credentials')
        authority_uri = AUTHORITY_HOST_URI + '/' + self.get_tenant_id()
        existing_cache = self.context.cache
        context = adal.AuthenticationContext(authority_uri, cache=existing_cache)
        new_token = context.acquire_token(credentials.token['resource'],
                                          credentials.token['user_id'],
                                          credentials.token['_client_id'])

        new_credentials = AADTokenCredentials(new_token, credentials.token.get('_client_id'))
        return new_credentials


class AzureAuthenticationStrategy(AuthenticationStrategy):

    def authenticate(self,
                     cli=None, user_account=None, user_account_browser=None,
                     service_principal=None, file_auth=None, msi=None,
                     tenant_id=None,
                     subscription_id=None,
                     client_id=None, client_secret=None,
                     username=None, password=None,
                     programmatic_execution=False,
                     **kargs):
        """
        Implements authentication for the Azure provider
        """
        try:

            # Set logging level to error for libraries as otherwise generates a lot of warnings
            logging.getLogger('adal-python').setLevel(logging.ERROR)
            logging.getLogger('msrest').setLevel(logging.ERROR)
            logging.getLogger('msrestazure.azure_active_directory').setLevel(logging.ERROR)
            logging.getLogger('urllib3').setLevel(logging.ERROR)
            logging.getLogger('cli.azure.cli.core').setLevel(logging.ERROR)

            context = None

            if cli:
                arm_credentials, subscription_id, tenant_id = \
                    get_azure_cli_credentials(with_tenant=True)
                aad_graph_credentials, placeholder_1, placeholder_2 = \
                    get_azure_cli_credentials(with_tenant=True, resource='https://graph.windows.net')

            elif user_account:

                if not (username and password):
                    if not programmatic_execution:
                        username = username if username else input("Username: ")
                        password = password if password else getpass("Password: ")
                    else:
                        raise AuthenticationException('Username and/or password not set')

                arm_credentials = UserPassCredentials(username, password)
                aad_graph_credentials = UserPassCredentials(username, password,
                                                            resource='https://graph.windows.net')

            elif user_account_browser:

                authority_uri = AUTHORITY_HOST_URI + '/' + tenant_id
                context = adal.AuthenticationContext(authority_uri, api_version=None)

                # Resource Manager
                resource_uri = 'https://management.core.windows.net/'
                code = context.acquire_user_code(resource_uri, AZURE_CLI_CLIENT_ID)
                print_info('To authenticate to the Resource Manager API, use a web browser to '
                           'access {} and enter the {} code.'.format(code['verification_url'],
                                                                     code['user_code']))
                arm_token = context.acquire_token_with_device_code(resource_uri, code, AZURE_CLI_CLIENT_ID)
                arm_credentials = AADTokenCredentials(arm_token, AZURE_CLI_CLIENT_ID)

                # AAD Graph
                resource_uri = 'https://graph.windows.net'
                code = context.acquire_user_code(resource_uri, AZURE_CLI_CLIENT_ID)
                print_info('To authenticate to the Azure AD Graph API, use a web browser to '
                           'access {} and enter the {} code.'.format(code['verification_url'],
                                                                     code['user_code']))
                aad_graph_token = context.acquire_token_with_device_code(resource_uri, code, AZURE_CLI_CLIENT_ID)
                aad_graph_credentials = AADTokenCredentials(aad_graph_token, AZURE_CLI_CLIENT_ID)

            elif service_principal:

                if not tenant_id:
                    if not programmatic_execution:
                        tenant_id = input("Tenant ID: ")
                    else:
                        raise AuthenticationException('No Tenant ID set')

                if not client_id:
                    if not programmatic_execution:
                        client_id = input("Client ID: ")
                    else:
                        raise AuthenticationException('No Client ID set')

                if not client_secret:
                    if not programmatic_execution:
                        client_secret = getpass("Client secret: ")
                    else:
                        raise AuthenticationException('No Client Secret set')

                arm_credentials = ServicePrincipalCredentials(
                    client_id=client_id,
                    secret=client_secret,
                    tenant=tenant_id
                )

                aad_graph_credentials = ServicePrincipalCredentials(
                    client_id=client_id,
                    secret=client_secret,
                    tenant=tenant_id,
                    resource='https://graph.windows.net'
                )

            elif file_auth:

                data = json.loads(file_auth.read())
                tenant_id = data.get('tenantId')
                client_id = data.get('clientId')
                client_secret = data.get('clientSecret')

                arm_credentials = ServicePrincipalCredentials(
                    client_id=client_id,
                    secret=client_secret,
                    tenant=tenant_id
                )

                aad_graph_credentials = ServicePrincipalCredentials(
                    client_id=client_id,
                    secret=client_secret,
                    tenant=tenant_id,
                    resource='https://graph.windows.net'
                )

            elif msi:

                arm_credentials = MSIAuthentication()
                aad_graph_credentials = MSIAuthentication(resource='https://graph.windows.net')

            else:
                raise AuthenticationException('Unknown authentication method')

            return AzureCredentials(arm_credentials,
                                    aad_graph_credentials,
                                    tenant_id, subscription_id,
                                    context)

        except Exception as e:
            if ', AdalError: Unsupported wstrust endpoint version. ' \
               'Current support version is wstrust2005 or wstrust13.' in e.args:
                raise AuthenticationException(
                    'You are likely authenticating with a Microsoft Account. '
                    'This authentication mode only support Azure Active Directory principal authentication.')

            raise AuthenticationException(e)



import logging

from oci.config import from_file
from oci.identity import IdentityClient

from ScoutSuite.providers.base.authentication_strategy import AuthenticationStrategy, AuthenticationException


class OracleCredentials:

    def __init__(self, config):
        self.config = config

    def get_scope(self):
        if 'compartment-id' in self.config:
            return self.config['compartment-id']
        else:
            return self.config['tenancy']


class OracleAuthenticationStrategy(AuthenticationStrategy):
    """
    Implements authentication for the AWS provider
    """

    def authenticate(self, profile=None, **kwargs):

        try:

            # Set logging level to error for libraries as otherwise generates a lot of warnings
            logging.getLogger('oci').setLevel(logging.ERROR)

            config = from_file(profile_name=profile)

            # Get the current user
            identity = IdentityClient(config)
            identity.get_user(config["user"]).data

            return OracleCredentials(config)

        except Exception as e:
            raise AuthenticationException(e)



from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.aws.facade.basefacade import AWSBaseFacade
from ScoutSuite.providers.aws.facade.utils import AWSFacadeUtils
from ScoutSuite.providers.utils import run_concurrently, map_concurrently


class CodeBuild(AWSBaseFacade):
    async def get_projects(self, region: str):
        codebuild_client = AWSFacadeUtils.get_client('codebuild', self.session, region)
        try:
            projects = await run_concurrently(lambda: codebuild_client.list_projects()['projects'])
        except Exception as e:
            print_exception(f'Failed to get CodeBuild projects: {e}')
            return []
        else:
            if not projects:
                return []
            return await map_concurrently(self._get_project_details, projects, region=region)
        
    async def _get_project_details(self, project: str, region: str):
        codebuild_client = AWSFacadeUtils.get_client('codebuild', self.session, region)
        try:
            project_details = await run_concurrently(lambda: codebuild_client.batch_get_projects(names=[project]))
        except Exception as e:
            print_exception(f'Failed to get CodeBuild project details: {e}')
            return {}
        else:
            project_details.pop('ResponseMetadata')
            project_details.pop('projectsNotFound')
            return project_details



from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureCompositeResources
from ScoutSuite.providers.azure.utils import get_resource_group_name
from ScoutSuite.providers.utils import get_non_provider_id

from .databases import Databases
from .server_azure_ad_administrators import ServerAzureAdAdministrators
from .server_blob_auditing_policies import ServerBlobAuditingPolicies
from .server_security_alert_policies import ServerSecurityAlertPolicies


class Servers(AzureCompositeResources):
    _children = [
        (Databases, 'databases'),
        (ServerAzureAdAdministrators, None),
        (ServerBlobAuditingPolicies, 'auditing'),
        (ServerSecurityAlertPolicies, 'threat_detection')
    ]

    def __init__(self, facade: AzureFacade, subscription_id: str):
        super().__init__(facade)
        self.subscription_id = subscription_id

    async def fetch_all(self):
        for raw_server in await self.facade.sqldatabase.get_servers(self.subscription_id):
            id, server = self._parse_server(raw_server)
            self[id] = server

        await self._fetch_children_of_all_resources(
            resources=self,
            scopes={server_id: {'resource_group_name': server['resource_group_name'],
                                'server_name': server['name'],
                                'subscription_id': self.subscription_id}
                    for (server_id, server) in self.items()}
        )

    def _parse_server(self, raw_server):
        server = {}
        server['id'] = get_non_provider_id(raw_server.id)
        server['name'] = raw_server.name
        server['resource_group_name'] = get_resource_group_name(raw_server.id)
        if raw_server.tags is not None:
            server['tags'] = ["{}:{}".format(key, value) for key, value in  raw_server.tags.items()]
        else:
            server['tags'] = []
        return server['id'], server



#!/usr/bin/env python

import os

scout_dir = 'ScoutSuite'
tests_dir = 'testsbase'

for root, dirnames, filenames in os.walk(scout_dir):
    for filename in filenames:
        if filename.startswith('__') or not filename.endswith('.py'):
            continue
        filepath = os.path.join(root, filename)
        tmp = filepath.split('.')[0].split('/')
        print(str(tmp))

        test = '# Import AWS utils\nfrom {} import *\n\n#\n# Test methods for {}\n#\n\nclass Test{}Class:\n\n'.format('.'.join(tmp), filepath, ''.join(t.title() for t in tmp))

        test_filename = 'test-%s.py' % '-'.join(tmp[1:])
        print(f'{filepath} --> {test_filename}')
        test_file = os.path.join(tests_dir, test_filename)
        if not os.path.isfile(test_file):
            with open(test_file, 'w+') as f:
                f.write(test)




from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.regions import Regions

from .trails import Trails


class CloudTrail(Regions):
    _children = [
        (Trails, 'trails')
    ]

    def __init__(self, facade: AWSFacade):
        super().__init__('cloudtrail', facade)

    async def finalize(self):
        global_events_logging = []

        for region in self['regions']:
            for trail_id, trail in self['regions'][region]['trails'].items():
                if 'HomeRegion' in trail and trail['HomeRegion'] != region:
                    # Part of a multi-region trail, skip until we find the whole object
                    continue
                if trail['IncludeGlobalServiceEvents'] and trail['IsLogging']:
                    global_events_logging.append((region, trail_id,))

        self['IncludeGlobalServiceEvents'] = len(global_events_logging) > 0
        self['DuplicatedGlobalServiceEvents'] = len(global_events_logging) > 1



import json

from botocore.exceptions import ClientError

from ScoutSuite.core.console import print_exception, print_debug
from ScoutSuite.providers.aws.facade.basefacade import AWSBaseFacade
from ScoutSuite.providers.aws.facade.utils import AWSFacadeUtils
from ScoutSuite.providers.utils import run_concurrently, get_and_set_concurrently


class S3Facade(AWSBaseFacade):
    async def get_buckets(self):
        try:
            # If there are regions specified, try for each of them until one works.
            # Otherwise, try all the available regions until one works.
            # This is required in case there's an IAM policy that denies access to APIs on a regional basis,
            # as per https://github.com/nccgroup/ScoutSuite/issues/727
            buckets = []
            exception = None
            region_list = self.regions if self.regions else await run_concurrently(lambda: self.session.get_available_regions('s3'))
            for region in region_list:
                try:
                    client = AWSFacadeUtils.get_client('s3', self.session, region)
                    buckets = await run_concurrently(lambda: client.list_buckets()['Buckets'])
                except Exception as e:
                    exception = e
                else:
                    exception = None  # Fix for https://github.com/nccgroup/ScoutSuite/issues/916#issuecomment-728783965
                    break
            if not buckets:
                if exception:
                    print_exception(f'Failed to list buckets: {exception}')
                return []
        except Exception as e:
            print_exception(f'Failed to list buckets: {e}')
            return []
        else:
            # We need first to retrieve bucket locations before retrieving bucket details
            await get_and_set_concurrently([self._get_and_set_s3_bucket_location], buckets, region=region)

            # Then we can retrieve bucket details concurrently
            await get_and_set_concurrently(
                [self._get_and_set_s3_bucket_logging,
                 self._get_and_set_s3_bucket_versioning,
                 self._get_and_set_s3_bucket_webhosting,
                 self._get_and_set_s3_bucket_default_encryption,
                 self._get_and_set_s3_acls,
                 self._get_and_set_s3_bucket_policy,
                 self._get_and_set_s3_bucket_tags,
                 self._get_and_set_s3_bucket_block_public_access],
                buckets)

            # Non-async post-processing
            for bucket in buckets:
                self._set_s3_bucket_secure_transport(bucket)
            # Try to update CreationDate of all buckets with the correct values from 'us-east-1'
            self._get_and_set_s3_bucket_creationdate(buckets)

            return buckets

    async def _get_and_set_s3_bucket_location(self, bucket: {}, region=None):
        client = AWSFacadeUtils.get_client('s3', self.session, region)
        try:
            location = await run_concurrently(lambda: client.get_bucket_location(Bucket=bucket['Name']))
        except Exception as e:
            print_exception('Failed to get bucket location for {}: {}'.format(bucket['Name'], e))
            location = None

        if location:
            region = location['LocationConstraint'] if location['LocationConstraint'] else 'us-east-1'

            # Fixes issue #59: location constraint can be either EU or eu-west-1 for Ireland...
            if region == 'EU':
                region = 'eu-west-1'
        else:
            region = None

        bucket['region'] = region

    async def _get_and_set_s3_bucket_logging(self, bucket: {}):
        client = AWSFacadeUtils.get_client('s3', self.session, bucket['region'], )
        try:
            logging = await run_concurrently(lambda: client.get_bucket_logging(Bucket=bucket['Name']))
        except Exception as e:
            print_exception('Failed to get logging configuration for {}: {}'.format(bucket['Name'], e))
            bucket['logging'] = 'Unknown'
        else:
            if 'LoggingEnabled' in logging:
                bucket['logging'] = \
                    logging['LoggingEnabled']['TargetBucket'] + '/' + logging['LoggingEnabled']['TargetPrefix']
            else:
                bucket['logging'] = 'Disabled'

    async def _get_and_set_s3_bucket_versioning(self, bucket: {}):
        client = AWSFacadeUtils.get_client('s3', self.session, bucket['region'])
        try:
            versioning = await run_concurrently(lambda: client.get_bucket_versioning(Bucket=bucket['Name']))
            bucket['versioning_status_enabled'] = self._status_to_bool(versioning.get('Status'))
            bucket['version_mfa_delete_enabled'] = self._status_to_bool(versioning.get('MFADelete'))
        except Exception as e:
            print_exception('Failed to get versioning configuration for {}: {}'.format(bucket['Name'], e))
            bucket['versioning_status_enabled'] = None
            bucket['version_mfa_delete_enabled'] = None

    async def _get_and_set_s3_bucket_webhosting(self, bucket: {}):
        client = AWSFacadeUtils.get_client('s3', self.session, bucket['region'])
        try:
            result = await run_concurrently(lambda: client.get_bucket_website(Bucket=bucket['Name']))
            bucket['web_hosting_enabled'] = 'IndexDocument' in result
        except Exception as e:
            if "NoSuchWebsiteConfiguration" in str(e):
                bucket['web_hosting_enabled'] = False
            else:
                print_exception('Failed to get web hosting configuration for {}: {}'.format(bucket['Name'], e))

    async def _get_and_set_s3_bucket_default_encryption(self, bucket: {}):
        bucket_name = bucket['Name']
        client = AWSFacadeUtils.get_client('s3', self.session, bucket['region'])
        try:
            config = await run_concurrently(lambda: client.get_bucket_encryption(Bucket=bucket['Name']))
            bucket['default_encryption_enabled'] = True
            bucket['default_encryption_algorithm'] = config.get('ServerSideEncryptionConfiguration', {})\
                .get('Rules', [{}])[0].get('ApplyServerSideEncryptionByDefault', {}).get('SSEAlgorithm')
            bucket['default_encryption_key'] = config.get('ServerSideEncryptionConfiguration', {})\
                .get('Rules', [{}])[0].get('ApplyServerSideEncryptionByDefault', {}).get('KMSMasterKeyID')
        except ClientError as e:
            if 'ServerSideEncryptionConfigurationNotFoundError' in e.response['Error']['Code']:
                bucket['default_encryption_enabled'] = False
                bucket['default_encryption_algorithm'] = None
                bucket['default_encryption_key'] = None
            else:
                bucket['default_encryption_enabled'] = None
                bucket['default_encryption_algorithm'] = None
                bucket['default_encryption_key'] = None
                print_exception(f'Failed to get encryption configuration for {bucket_name}: {e}')
        except Exception as e:
            bucket['default_encryption'] = 'Unknown'
            bucket['default_encryption_algorithm'] = None
            bucket['default_encryption_key'] = None
            print_exception(f'Failed to get encryption configuration for {bucket_name}: {e}')

    async def _get_and_set_s3_acls(self, bucket: {}, key_name=None):
        bucket_name = bucket['Name']
        client = AWSFacadeUtils.get_client('s3', self.session, bucket['region'])
        try:
            grantees = {}
            if key_name:
                grants = await run_concurrently(lambda: client.get_object_acl(Bucket=bucket_name, Key=key_name))
            else:
                grants = await run_concurrently(lambda: client.get_bucket_acl(Bucket=bucket_name))
            for grant in grants['Grants']:
                if 'ID' in grant['Grantee']:
                    grantee = grant['Grantee']['ID']
                    display_name = grant['Grantee']['DisplayName'] if \
                        'DisplayName' in grant['Grantee'] else grant['Grantee']['ID']
                elif 'URI' in grant['Grantee']:
                    grantee = grant['Grantee']['URI'].split('/')[-1]
                    display_name = self._s3_group_to_string(grant['Grantee']['URI'])
                else:
                    grantee = display_name = 'Unknown'
                permission = grant['Permission']
                grantees.setdefault(grantee, {})
                grantees[grantee]['DisplayName'] = display_name
                if 'URI' in grant['Grantee']:
                    grantees[grantee]['URI'] = grant['Grantee']['URI']
                grantees[grantee].setdefault('permissions', self._init_s3_permissions())
                self._set_s3_permissions(grantees[grantee]['permissions'], permission)
            bucket['grantees'] = grantees
        except Exception as e:
            print_exception(f'Failed to get ACL configuration for {bucket_name}: {e}')
            bucket['grantees'] = {}

    async def _get_and_set_s3_bucket_policy(self, bucket: {}):
        client = AWSFacadeUtils.get_client('s3', self.session, bucket['region'])
        try:
            bucket_policy = await run_concurrently(lambda: client.get_bucket_policy(Bucket=bucket['Name']))
            bucket['policy'] = json.loads(bucket_policy['Policy'])
        except ClientError as e:
            if e.response['Error']['Code'] != 'NoSuchBucketPolicy':
                print_exception('Failed to get bucket policy for {}: {}'.format(bucket['Name'], e))
        except Exception as e:
            print_exception('Failed to get bucket policy for {}: {}'.format(bucket['Name'], e))
            bucket['grantees'] = {}

    async def _get_and_set_s3_bucket_tags(self, bucket: {}):
        client = AWSFacadeUtils.get_client('s3', self.session, bucket['region'])
        try:
            bucket_tagset = await run_concurrently(lambda: client.get_bucket_tagging(Bucket=bucket['Name']))
            bucket['tags'] = {x['Key']: x['Value'] for x in bucket_tagset['TagSet']}
        except ClientError as e:
            if e.response['Error']['Code'] != 'NoSuchTagSet':
                print_exception('Failed to get bucket tags for {}: {}'.format(bucket['Name'], e))
        except Exception as e:
            print_exception('Failed to get bucket tags for {}: {}'.format(bucket['Name'], e))
            bucket['tags'] = {}

    async def _get_and_set_s3_bucket_block_public_access(self, bucket: {}):
        client = AWSFacadeUtils.get_client('s3', self.session, bucket['region'])
        try:
            bucket_public_access_block_conf = await run_concurrently(lambda: client.get_public_access_block(Bucket=bucket['Name']))
            bucket['public_access_block_configuration'] = bucket_public_access_block_conf['PublicAccessBlockConfiguration']
        except ClientError as e:
            # No such configuration found for the bucket, nothing to be done
            pass
        except Exception as e:
            print_exception('Failed to get the public access block configuration for {}: {}'.format(bucket['Name'], e))

    def _get_and_set_s3_bucket_creationdate(self, buckets):
        # When using region other than 'us-east-1', the 'CreationDate' is the last modified time according to bucket's
        # last replication in the respective region
        # Source: https://github.com/aws/aws-cli/issues/3597#issuecomment-424167129
        # Fixes issue https://github.com/nccgroup/ScoutSuite/issues/858
        client = AWSFacadeUtils.get_client('s3', self.session, 'us-east-1')
        try:
            buckets_useast1 = client.list_buckets()['Buckets']
            for bucket in buckets:
                # Find the bucket with the same name and update 'CreationDate' from the 'us-east-1' region data,
                # if doesn't exist keep the original value
                bucket['CreationDate'] = next((b['CreationDate'] for b in buckets_useast1 if
                                               b['Name'] == bucket['Name']), bucket['CreationDate'])
        except Exception as e:
            # Only output exception when in debug mode
            print_debug('Failed to get bucket creation date from "us-east-1" region')

    def _set_s3_bucket_secure_transport(self, bucket: {}):
        try:
            if 'policy' in bucket:
                bucket['secure_transport_enabled'] = False
                for statement in bucket['policy']['Statement']:
                    # evaluate statement to see if it contains a condition disallowing HTTP transport
                    # TODO this might not cover all cases
                    if 'Condition' in statement and \
                            'Bool' in statement['Condition'] and \
                            'aws:SecureTransport' in statement['Condition']['Bool'] and \
                            ((statement['Condition']['Bool']['aws:SecureTransport'] == 'false' and
                              statement['Effect'] == 'Deny') or
                             (statement['Condition']['Bool']['aws:SecureTransport'] == 'true' and
                              statement['Effect'] == 'Allow')):
                        bucket['secure_transport_enabled'] = True
            else:
                bucket['secure_transport_enabled'] = False
        except Exception as e:
            print_exception('Failed to evaluate bucket policy for {}: {}'.format(bucket['Name'], e))
            bucket['secure_transport'] = None

    def get_s3_public_access_block(self, account_id):
        # We need a region to generate the client
        # However, the settings are global, so they are not region-dependent
        region = 'us-east-1'
        client = AWSFacadeUtils.get_client('s3control', self.session, region)
        try:
            s3_public_access_block = client.get_public_access_block(AccountId=account_id)
            return s3_public_access_block['PublicAccessBlockConfiguration']
        except ClientError:
            # No public access block configuration at the S3 level, returning the default
            return {
                "BlockPublicAcls": False,
                "IgnorePublicAcls": False,
                "BlockPublicPolicy": False,
                "RestrictPublicBuckets": False
            }
        except Exception as e:
            print_exception(
                f'Failed to get the public access block configuration for the account {account_id}: {e}')
            return None

    @staticmethod
    def _init_s3_permissions():
        permissions = {'read': False, 'write': False, 'read_acp': False, 'write_acp': False}
        return permissions

    @staticmethod
    def _set_s3_permissions(permissions: str, name: str):
        if name == 'READ' or name == 'FULL_CONTROL':
            permissions['read'] = True
        if name == 'WRITE' or name == 'FULL_CONTROL':
            permissions['write'] = True
        if name == 'READ_ACP' or name == 'FULL_CONTROL':
            permissions['read_acp'] = True
        if name == 'WRITE_ACP' or name == 'FULL_CONTROL':
            permissions['write_acp'] = True

    @staticmethod
    def _s3_group_to_string(uri: str):
        if uri == 'http://acs.amazonaws.com/groups/global/AuthenticatedUsers':
            return 'Authenticated users'
        elif uri == 'http://acs.amazonaws.com/groups/global/AllUsers':
            return 'Everyone'
        elif uri == 'http://acs.amazonaws.com/groups/s3/LogDelivery':
            return 'Log delivery'
        else:
            return uri

    @staticmethod
    def _status_to_bool(value: str):
        """ Converts a string to True if it is equal to 'Enabled' or to False otherwise. """
        return value == 'Enabled'



import json

from ScoutSuite.core.console import print_exception, print_info, print_debug, print_error
from ScoutSuite.providers.gcp.facade.basefacade import GCPBaseFacade
from ScoutSuite.providers.gcp.facade.cloudresourcemanager import CloudResourceManagerFacade
from ScoutSuite.providers.gcp.facade.cloudsql import CloudSQLFacade
from ScoutSuite.providers.gcp.facade.memorystoreredis import MemoryStoreRedisFacade
from ScoutSuite.providers.gcp.facade.cloudstorage import CloudStorageFacade
from ScoutSuite.providers.gcp.facade.gce import GCEFacade
from ScoutSuite.providers.gcp.facade.dns import DNSFacade
from ScoutSuite.providers.gcp.facade.iam import IAMFacade
from ScoutSuite.providers.gcp.facade.kms import KMSFacade
from ScoutSuite.providers.gcp.facade.stackdriverlogging import StackdriverLoggingFacade
from ScoutSuite.providers.gcp.facade.stackdrivermonitoring import StackdriverMonitoringFacade
from ScoutSuite.providers.gcp.facade.gke import GKEFacade
from ScoutSuite.providers.gcp.facade.utils import GCPFacadeUtils
from ScoutSuite.utils import format_service_name


class GCPFacade(GCPBaseFacade):
    def __init__(self,
                 default_project_id=None, project_id=None, folder_id=None, organization_id=None, all_projects=None):
        super().__init__('cloudresourcemanager', 'v1')

        self.default_project_id = default_project_id
        self.all_projects = all_projects
        self.project_id = project_id
        self.folder_id = folder_id
        self.organization_id = organization_id

        self.cloudresourcemanager = CloudResourceManagerFacade()
        self.cloudsql = CloudSQLFacade()
        self.cloudstorage = CloudStorageFacade()
        self.memorystoreredis = MemoryStoreRedisFacade()
        self.gce = GCEFacade()
        self.iam = IAMFacade()
        self.kms = KMSFacade()
        self.dns = DNSFacade()
        self.stackdriverlogging = StackdriverLoggingFacade()
        self.stackdrivermonitoring = StackdriverMonitoringFacade()

        # Instantiate facades for proprietary services
        try:
            self.gke = GKEFacade(self.gce)
        except NameError as _:
            pass

    async def get_projects(self):
        try:

            # All projects to which the user / Service Account has access to
            if self.all_projects:
                return await self._get_projects_recursively(
                    parent_type='all', parent_id=None)
            # Project passed through the CLI
            elif self.project_id:
                return await self._get_projects_recursively(
                    parent_type='project', parent_id=self.project_id)
            # Folder passed through the CLI
            elif self.folder_id:
                return await self._get_projects_recursively(
                    parent_type='folder', parent_id=self.folder_id)
            # Organization passed through the CLI
            elif self.organization_id:
                return await self._get_projects_recursively(
                    parent_type='organization', parent_id=self.organization_id)
            # Project inferred from default configuration
            elif self.default_project_id:
                return await self._get_projects_recursively(
                    parent_type='project', parent_id=self.default_project_id)
            # Raise exception if none of the above
            else:
                print_info(
                    "Could not infer the Projects to scan and no default Project ID was found.")
                return []

        except Exception as e:
            print_exception(f'Failed to retrieve projects: {e}')
            return []

    async def _get_projects_recursively(self, parent_type, parent_id):
        """
        Returns all the projects in a given organization or folder. For a project_id it only returns the project
        details.

        # FIXME can't currently be done with API client library as it consumes v1 which doesn't support folders
        resource_manager_client = resource_manager.Client(credentials=self.credentials)
        project_list = resource_manager_client.list_projects()
        for p in project_list:
            if p.parent['id'] == self.organization_id and p.status == 'ACTIVE':
                projects.append(p.project_id)
        """

        if parent_type not in ['project', 'organization', 'folder', 'all']:
            return None

        resourcemanager_client = self._get_client()
        resourcemanager_client_v2 = self._build_arbitrary_client('cloudresourcemanager', 'v2', force_new=True)

        projects = []

        try:
            projects_group = resourcemanager_client.projects()

            if parent_type == 'project':
                request = resourcemanager_client.projects().list(filter='id:"%s"' % parent_id)
            elif parent_type == 'all':
                request = resourcemanager_client.projects().list()
            # get parent children projects
            else:
                request = resourcemanager_client.projects().list(filter='parent.id:"%s"' % parent_id)

                # get parent children projects in children folders recursively
                folder_request = resourcemanager_client_v2.folders().list(parent=f'{parent_type}s/{parent_id}')
                folder_response = await GCPFacadeUtils.get_all('folders', folder_request, projects_group)
                for folder in folder_response:
                    projects.extend(await self._get_projects_recursively("folder", folder['name'].strip('folders/')))

            project_response = await GCPFacadeUtils.get_all('projects', request, projects_group)
            if project_response:
                for project in project_response:
                    if project['lifecycleState'] == "ACTIVE":
                        projects.append(project)
            else:
                print_exception('No Projects Found, '
                                'you may have specified a non-existing Organization, Folder or Project')

        except Exception as e:
            try:
                content = e.content.decode("utf-8")
                content_dict = json.loads(content)
                print_exception(f'Unable to list accessible Projects: {content_dict.get("error").get("message")}')
            except Exception as e:
                print_exception(f'Unable to list accessible Projects: {e}')

        finally:
            return projects

    async def is_api_enabled(self, project_id, service):
        """
        Given a project ID and service name, this method tries to determine if the service's API is enabled
        """

        # All projects have IAM policies regardless of whether the IAM API is enabled.
        if service == 'IAM':
            return True

        serviceusage_client = self._build_arbitrary_client('serviceusage', 'v1', force_new=True)
        services = serviceusage_client.services()
        try:
            request = services.list(parent=f'projects/{project_id}')
            services_response = await GCPFacadeUtils.get_all('services', request, services)
        except Exception as e:
            print_exception(f'Could not fetch the state of services for project \"{project_id}\", '
                            f'including {format_service_name(service.lower())} in the execution', {'exception': e})
            return True

        # These are hardcoded endpoint correspondences as there's no easy way to do this.
        if service == 'KMS':
            endpoint = 'cloudkms'
        elif service == 'CloudStorage':
            endpoint = 'storage-component'
        elif service == 'CloudSQL':
            endpoint = 'sql-component'
        elif service == 'ComputeEngine':
            endpoint = 'compute'
        elif service == 'KubernetesEngine':
            endpoint = 'container'
        elif service == 'StackdriverLogging':
            endpoint = 'logging'
        elif service == 'StackdriverMonitoring':
            endpoint = 'monitoring'
        elif service == 'MemoryStore':
            endpoint = 'redis'
        elif service =='DNS':
            endpoint='dns'
        else:
            print_debug('Could not validate the state of the {} API for project \"{}\", '
                        'including it in the execution'.format(format_service_name(service.lower()), project_id))
            return True

        for s in services_response:
            if endpoint in s.get('name'):
                if s.get('state') == 'ENABLED':
                    return True
                else:
                    print_info('{} API not enabled for project \"{}\", skipping'.format(format_service_name(service.lower()),
                                                                                        project_id))
                    return False

        print_error(f'Could not validate the state of the {format_service_name(service.lower())} API '
                    f'for project \"{project_id}\", including it in the execution')
        return True



from azure.mgmt.authorization import AuthorizationManagementClient

from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.utils import run_concurrently
from ScoutSuite.utils import get_user_agent


class RBACFacade:

    def __init__(self, credentials):
        self.credentials = credentials

    def get_client(self, subscription_id: str):
        client = AuthorizationManagementClient(self.credentials.get_credentials('arm'),
                                             subscription_id=subscription_id)
        client._client.config.add_user_agent(get_user_agent())
        return client

    async def get_roles(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            scope = f'/subscriptions/{subscription_id}'
            return await run_concurrently(lambda: list(client.role_definitions.list(scope=scope)))
        except Exception as e:
            print_exception(f'Failed to retrieve roles: {e}')
            return []

    async def get_role_assignments(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(lambda: list(client.role_assignments.list()))
        except Exception as e:
            print_exception(f'Failed to retrieve role assignments: {e}')
            return []



from ScoutSuite.providers.aws.resources.base import AWSCompositeResources
from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.cloudfront.distributions import Distributions

from .distributions import Distributions


class CloudFront(AWSCompositeResources):
    _children = [
        (Distributions, 'distributions')
    ]

    def __init__(self, facade: AWSFacade):
        super(CloudFront, self).__init__(facade)
        self.service = 'cloudfront'

    async def fetch_all(self, partition_name='aws', **kwargs):
        await self._fetch_children(self)



from azure.graphrbac import GraphRbacManagementClient

from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.utils import run_concurrently
from ScoutSuite.utils import get_user_agent


class AADFacade:

    def __init__(self, credentials):
        self.credentials = credentials

    def get_client(self):
        client = GraphRbacManagementClient(self.credentials.get_credentials('aad_graph'),
                                         tenant_id=self.credentials.get_tenant_id())
        client._client.config.add_user_agent(get_user_agent())
        return client

    async def get_users(self):
        try:
            # This filters down the users which are pulled from the directory, otherwise for large tenants this
            # gets out of hands.
            # See https://github.com/nccgroup/ScoutSuite/issues/698
            user_filter = " and ".join([
                'userType eq \'Guest\''
            ])
            return await run_concurrently(lambda: list(self.get_client().users.list(filter=user_filter)))
        except Exception as e:
            print_exception(f'Failed to retrieve users: {e}')
            return []

    async def get_user(self, user_id):
        try:
            return await run_concurrently(lambda: self.get_client().users.get(user_id))
        except Exception as e:
            print_exception(f'Failed to retrieve user {user_id}: {e}')
            return None

    async def get_groups(self):
        try:
            return await run_concurrently(lambda: list(self.get_client().groups.list()))
        except Exception as e:
            print_exception(f'Failed to retrieve groups: {e}')
            return []

    async def get_user_groups(self, user_id):
        try:
            return await run_concurrently(lambda: list(
                self.get_client().users.get_member_groups(object_id=user_id,
                                                          security_enabled_only=False))
                                          )
        except Exception as e:
            print_exception(f'Failed to retrieve user\'s groups: {e}')
            return []

    async def get_service_principals(self):
        try:
            return await run_concurrently(lambda: list(self.get_client().service_principals.list()))
        except Exception as e:
            print_exception(f'Failed to retrieve service principals: {e}')
            return []

    async def get_applications(self):
        try:
            return await run_concurrently(lambda: list(self.get_client().applications.list()))
        except Exception as e:
            print_exception(f'Failed to retrieve applications: {e}')
            return []



from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.gcp.facade.basefacade import GCPBaseFacade
from ScoutSuite.providers.gcp.facade.utils import GCPFacadeUtils
from ScoutSuite.providers.utils import run_concurrently


class DNSFacade(GCPBaseFacade):
    def __init__(self):
        super().__init__('dns', 'v1')

    async def get_zones(self, project_id):
        try:
            dns_client = self._get_client()
            return await run_concurrently(
                lambda: dns_client.managedZones().list(project=project_id).execute()
            )
        except Exception as e:
            print_exception(f'Failed to retrieve zones: {e}')
            return []



import json

from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.aws.facade.basefacade import AWSBaseFacade
from ScoutSuite.providers.aws.facade.utils import AWSFacadeUtils


class LambdaFacade(AWSBaseFacade):
    async def get_functions(self, region):
        try:
            return await AWSFacadeUtils.get_all_pages('lambda', region, self.session, 'list_functions', 'Functions')
        except Exception as e:
            print_exception(f'Failed to get Lambda functions: {e}')
            return []

    async def get_access_policy(self, function_name, region):
        client = AWSFacadeUtils.get_client('lambda', self.session, region)
        try:
            policy = client.get_policy(FunctionName=function_name)
            if policy is not None and 'Policy' in policy:
                return json.loads(policy['Policy'])
        except Exception as e:
            # If there's no policy, it will return this exception. Hence why we ignore.
            if "ResourceNotFoundException" not in str(e):
                print_exception('Failed to get Lambda access policy: {}'.format(e))
            return None

    async def get_role_with_managed_policies(self, role_name):
        client = AWSFacadeUtils.get_client('iam', self.session)
        try:
            role = client.get_role(RoleName=role_name)['Role']
            managed_policies = client.list_attached_role_policies(RoleName=role_name)['AttachedPolicies']
            for policy in managed_policies:
                policy_version = client.get_policy(PolicyArn=policy['PolicyArn'])
                if 'Policy' in policy_version and 'DefaultVersionId' in policy_version['Policy']:
                    policy_version = policy_version['Policy']['DefaultVersionId']
                    document = client.get_policy_version(PolicyArn=policy['PolicyArn'], VersionId=policy_version)
                    if 'PolicyVersion' in document and 'Document' in document['PolicyVersion']:
                        policy['Document'] = document['PolicyVersion']['Document']
            role['policies'] = managed_policies
            return role
        except Exception as e:
            print_exception('Failed to get role from managed policies: {}'.format(e))
            return None

    async def get_env_variables(self, function_name, region):
        client = AWSFacadeUtils.get_client('lambda', self.session, region)
        try:
            function_configuration = client.get_function_configuration(FunctionName=function_name)
            if "Environment" in function_configuration and "Variables" in function_configuration["Environment"]:
                return function_configuration["Environment"]["Variables"]
        except Exception as e:
            print_exception('Failed to get Lambda function configuration: {}'.format(e))
        return []




from ScoutSuite.providers.azure.resources.subscriptions import Subscriptions

from .instances import Instances
from .disks import Disks
from .snapshots import Snapshots
from .images import Images


class VirtualMachines(Subscriptions):
    _children = [
        (Instances, 'instances'),
        (Disks, 'disks'),
        (Snapshots, 'snapshots'),
        (Images, 'images'),
    ]



from ScoutSuite.providers.azure.resources.subscriptions import Subscriptions

from .storage_accounts import StorageAccounts


class StorageAccounts(Subscriptions):
    _children = [
        (StorageAccounts, 'storage_accounts')
    ]




from boto3.session import Session

from ScoutSuite.providers.aws.facade.acm import AcmFacade
from ScoutSuite.providers.aws.facade.awslambda import LambdaFacade
from ScoutSuite.providers.aws.facade.basefacade import AWSBaseFacade
from ScoutSuite.providers.aws.facade.cloudformation import CloudFormation
from ScoutSuite.providers.aws.facade.cloudtrail import CloudTrailFacade
from ScoutSuite.providers.aws.facade.cloudwatch import CloudWatch
from ScoutSuite.providers.aws.facade.cloudfront import CloudFront
from ScoutSuite.providers.aws.facade.codebuild import CodeBuild
from ScoutSuite.providers.aws.facade.config import ConfigFacade
from ScoutSuite.providers.aws.facade.directconnect import DirectConnectFacade
from ScoutSuite.providers.aws.facade.dynamodb import DynamoDBFacade
from ScoutSuite.providers.aws.facade.ec2 import EC2Facade
from ScoutSuite.providers.aws.facade.efs import EFSFacade
from ScoutSuite.providers.aws.facade.elasticache import ElastiCacheFacade
from ScoutSuite.providers.aws.facade.elb import ELBFacade
from ScoutSuite.providers.aws.facade.elbv2 import ELBv2Facade
from ScoutSuite.providers.aws.facade.emr import EMRFacade
from ScoutSuite.providers.aws.facade.iam import IAMFacade
from ScoutSuite.providers.aws.facade.kms import KMSFacade
from ScoutSuite.providers.aws.facade.rds import RDSFacade
from ScoutSuite.providers.aws.facade.redshift import RedshiftFacade
from ScoutSuite.providers.aws.facade.route53 import Route53Facade
from ScoutSuite.providers.aws.facade.s3 import S3Facade
from ScoutSuite.providers.aws.facade.ses import SESFacade
from ScoutSuite.providers.aws.facade.sns import SNSFacade
from ScoutSuite.providers.aws.facade.sqs import SQSFacade
from ScoutSuite.providers.aws.facade.secretsmanager import SecretsManagerFacade
from ScoutSuite.providers.aws.utils import get_aws_account_id, get_partition_name
from ScoutSuite.providers.utils import run_concurrently

from ScoutSuite.core.conditions import print_error

# Try to import proprietary facades
try:
    from ScoutSuite.providers.aws.facade.cognito_private import CognitoFacade
except ImportError:
    pass
try:
    from ScoutSuite.providers.aws.facade.docdb_private import DocDBFacade
except ImportError:
    pass
try:
    from ScoutSuite.providers.aws.facade.ecs_private import ECSFacade
except ImportError:
    pass
try:
    from ScoutSuite.providers.aws.facade.ecr_private import ECRFacade
except ImportError:
    pass
try:
    from ScoutSuite.providers.aws.facade.eks_private import EKSFacade
except ImportError:
    pass
try:
    from ScoutSuite.providers.aws.facade.guardduty_private import GuardDutyFacade
except ImportError:
    pass
try:
    from ScoutSuite.providers.aws.facade.ssm_private import SSMFacade
except ImportError:
    pass


class AWSFacade(AWSBaseFacade):
    def __init__(self, credentials=None):
        super().__init__()
        self.owner_id = get_aws_account_id(credentials.session)
        self.partition = get_partition_name(credentials.session)
        self.session = credentials.session
        self._instantiate_facades()

    async def build_region_list(self, service: str, chosen_regions=None, excluded_regions=None, partition_name='aws'):

        available_services = None
        try:
            available_services = await run_concurrently(
                lambda: Session(region_name='us-east-1').get_available_services())
        except Exception as e:
            # see https://github.com/nccgroup/ScoutSuite/issues/548
            # If failed with the us-east-1 region, we'll try to use the region from the profile
            try:
                available_services = await run_concurrently(
                    lambda: Session(region_name=self.session.region_name).get_available_services())
            except Exception as e:
                # see https://github.com/nccgroup/ScoutSuite/issues/685
                # If above failed, and regions were explicitly specified, will try with those until one works
                if chosen_regions:
                    for region in chosen_regions:
                        try:
                            available_services = await run_concurrently(
                                lambda: Session(region_name=region).get_available_services())
                            break
                        except Exception as e:
                            exception = e
                    if not available_services:
                        raise exception
                else:
                    raise e

        if service not in available_services:
            # the cognito service is a composition of two boto3 services
            if service == "cognito":
                if "cognito-idp" not in available_services:
                    raise Exception('Service cognito-idp is not available.')
                elif "cognito-identity" not in available_services:
                    raise Exception('Service cognito-identity is not available.')
            else:
                raise Exception('Service ' + service + ' is not available.')

        regions = None
        try:
            # the cognito service is a composition of two boto3 services
            if service != "cognito":
                regions = await run_concurrently(
                    lambda: Session(region_name='us-east-1').get_available_regions(service,
                                                                                   partition_name))
            else:
                idp_regions = await run_concurrently(
                    lambda: Session(region_name='us-east-1').get_available_regions("cognito-idp",
                                                                                   partition_name))
                identity_regions = await run_concurrently(
                    lambda: Session(region_name='us-east-1').get_available_regions("cognito-identity",
                                                                                   partition_name))
                regions = [value for value in idp_regions if value in identity_regions]
        except Exception as e:
            # see https://github.com/nccgroup/ScoutSuite/issues/548
            # If failed with the us-east-1 region, we'll try to use the region from the profile
            try:
                # the cognito service is a composition of two boto3 services
                if service != "cognito":
                    regions = await run_concurrently(
                        lambda: Session(region_name=self.session.region_name).get_available_regions(service,
                                                                                                    partition_name))
                else:
                    idp_regions = await run_concurrently(
                        lambda: Session(region_name=self.session.region_name).get_available_regions("cognito-idp",
                                                                                                    partition_name))
                    identity_regions = await run_concurrently(
                        lambda: Session(region_name=self.session.region_name).get_available_regions("cognito-identity",
                                                                                                    partition_name))
                    regions = [value for value in idp_regions if value in identity_regions]
            except Exception as e:
                # see https://github.com/nccgroup/ScoutSuite/issues/685
                # If above failed, and regions were explicitly specified, will try with those until one works
                if chosen_regions:
                    for region in chosen_regions:
                        try:
                            # the cognito service is a composition of two boto3 services
                            if service != "cognito":
                                regions = await run_concurrently(
                                    lambda: Session(region_name=region).get_available_regions(service,
                                                                                              partition_name))
                            else:
                                idp_regions = await run_concurrently(
                                    lambda: Session(region_name=region).get_available_regions(
                                        "cognito-idp",
                                        partition_name))
                                identity_regions = await run_concurrently(
                                    lambda: Session(region_name=region).get_available_regions(
                                        "cognito-identity",
                                        partition_name))
                                regions = [value for value in idp_regions if value in identity_regions]
                            break
                        except Exception as e:
                            exception = e
                    if not regions:
                        raise exception
                else:
                    raise e

        if not regions:
            # Could be an instance of https://github.com/boto/boto3/issues/1662
            if service == 'eks':  # TODO fix when the issue is resolved
                regions = ['ap-east-1',
                           'ap-northeast-1',
                           'ap-northeast-2',
                           'ap-south-1',
                           'ap-southeast-1',
                           'ap-southeast-2',
                           'ca-central-1',
                           'eu-central-1',
                           'eu-north-1',
                           'eu-west-1',
                           'eu-west-2',
                           'eu-west-3',
                           'me-south-1',
                           'sa-east-1',
                           'us-east-1',
                           'us-east-2',
                           # 'us-west-1',
                           'us-west-2']
            else:
                print_error('"get_available_regions" returned an empty array for service "{}", '
                            'something is wrong'.format(service))

        # identify regions that are not opted-in
        ec2_not_opted_in_regions = None
        try:
            ec2_not_opted_in_regions = self.session.client('ec2', 'us-east-1') \
                .describe_regions(AllRegions=True, Filters=[{'Name': 'opt-in-status', 'Values': ['not-opted-in']}])
        except Exception as e:
            # see https://github.com/nccgroup/ScoutSuite/issues/548
            # If failed with the us-east-1 region, we'll try to use the region from the profile
            try:
                ec2_not_opted_in_regions = \
                    self.session.client('ec2', self.session.region_name). \
                        describe_regions(AllRegions=True,
                                         Filters=[{'Name': 'opt-in-status',
                                                   'Values': ['not-opted-in']}])
            except Exception as e:
                # see https://github.com/nccgroup/ScoutSuite/issues/685
                # If above failed, and regions were explicitly specified, will try with those until
                # one works
                if chosen_regions:
                    for region in chosen_regions:
                        try:
                            ec2_not_opted_in_regions = self.session.client('ec2', region).describe_regions(
                                AllRegions=True,
                                Filters=[{'Name': 'opt-in-status',
                                          'Values': ['not-opted-in']}])
                            break
                        except Exception as e:
                            exception = e
                    if not ec2_not_opted_in_regions:
                        raise exception
                else:
                    raise e

        not_opted_in_regions = []
        if ec2_not_opted_in_regions['Regions']:
            for r in ec2_not_opted_in_regions['Regions']:
                not_opted_in_regions.append(r['RegionName'])

        # include specific regions
        if chosen_regions:
            regions = [r for r in regions if r in chosen_regions]
        # exclude specific regions
        if excluded_regions:
            regions = [r for r in regions if r not in excluded_regions]
        # exclude not opted in regions
        if not_opted_in_regions:
            regions = [r for r in regions if r not in not_opted_in_regions]

        return regions

    def _instantiate_facades(self):
        self.ec2 = EC2Facade(self.session, self.owner_id)
        self.acm = AcmFacade(self.session)
        self.awslambda = LambdaFacade(self.session)
        self.cloudformation = CloudFormation(self.session)
        self.cloudtrail = CloudTrailFacade(self.session)
        self.cloudwatch = CloudWatch(self.session)
        self.config = ConfigFacade(self.session)
        self.directconnect = DirectConnectFacade(self.session)
        self.dynamodb = DynamoDBFacade(self.session)
        self.efs = EFSFacade(self.session)
        self.elasticache = ElastiCacheFacade(self.session)
        self.route53 = Route53Facade(self.session)
        self.cloudfront = CloudFront(self.session)
        self.codebuild = CodeBuild(self.session)
        self.elb = ELBFacade(self.session)
        self.elbv2 = ELBv2Facade(self.session)
        self.iam = IAMFacade(self.session)
        self.kms = KMSFacade(self.session)
        self.rds = RDSFacade(self.session)
        self.redshift = RedshiftFacade(self.session)
        self.s3 = S3Facade(self.session)
        self.ses = SESFacade(self.session)
        self.sns = SNSFacade(self.session)
        self.sqs = SQSFacade(self.session)
        self.secretsmanager = SecretsManagerFacade(self.session)
        self.emr = EMRFacade(self.session)

        # Instantiate facades for proprietary services
        try:
            self.cognito = CognitoFacade(self.session)
        except NameError:
            pass
        try:
            self.docdb = DocDBFacade(self.session)
        except NameError:
            pass
        try:
            self.ecs = ECSFacade(self.session)
        except NameError:
            pass
        try:
            self.ecr = ECRFacade(self.session)
        except NameError:
            pass
        try:
            self.eks = EKSFacade(self.session)
        except NameError:
            pass
        try:
            self.guardduty = GuardDutyFacade(self.session)
        except NameError:
            pass
        try:
            self.ssm = SSMFacade(self.session)
        except NameError:
            pass



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.base import AWSResources
from ScoutSuite.providers.aws.utils import get_name, format_arn


class FlowLogs(AWSResources):
    def __init__(self, facade: AWSFacade, region: str):
        super().__init__(facade)
        self.facade = facade
        self.region = region
        self.partition = facade.partition
        self.service = 'vpc'
        self.resource_type = 'flow-log'

    async def fetch_all(self):
        raw_logs = await self.facade.ec2.get_flow_logs(self.region)

        for raw_log in raw_logs:
            id, log = self._parse_log(raw_log)
            self[id] = log

    def _parse_log(self, raw_flow_log):
        flow_log_dict = {}
        flow_log_dict['name'] = flow_log_dict['id'] = raw_flow_log.get('FlowLogId')
        flow_log_dict['creation_time'] = raw_flow_log.get('CreationTime')
        flow_log_dict['deliver_logs_error_message'] = raw_flow_log.get('DeliverLogsErrorMessage')
        flow_log_dict['deliver_logs_status'] = raw_flow_log.get('DeliverLogsStatus')
        flow_log_dict['flow_log_status'] = raw_flow_log.get('FlowLogStatus')
        flow_log_dict['resource_id'] = raw_flow_log.get('ResourceId')
        flow_log_dict['traffic_type'] = raw_flow_log.get('TrafficType')
        flow_log_dict['log_destination_type'] = raw_flow_log.get('LogDestinationType')
        flow_log_dict['log_destination'] = raw_flow_log.get('LogDestination')
        flow_log_dict['log_format'] = raw_flow_log.get('LogFormat')
        flow_log_dict['tags'] = raw_flow_log.get('Tags')
        flow_log_dict['max_aggregation_interval'] = raw_flow_log.get('MaxAggregationInterval')
        flow_log_dict['arn'] = format_arn(self.partition, self.service, self.region, '', raw_flow_log.get('FlowLogId'), self.resource_type)
        return flow_log_dict['id'], flow_log_dict




from google.cloud import kms
from google.api_core.gapic_v1.client_info import ClientInfo

from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.gcp.facade.basefacade import GCPBaseFacade
from ScoutSuite.providers.gcp.facade.utils import GCPFacadeUtils
from ScoutSuite.providers.utils import run_concurrently
from ScoutSuite.utils import get_user_agent


class KMSFacade(GCPBaseFacade):
    def __init__(self):
        # This facade is currently using both libraries as the Cloud Client library doesn't support locations

        # Cloud Client
        client_info = ClientInfo(user_agent=get_user_agent())
        self.cloud_client = kms.KeyManagementServiceClient(client_info=client_info)
        # self.cloud_client = kms.KeyManagementServiceClient()

        super().__init__('cloudkms', 'v1')  # API Client

    async def get_locations(self, project_id: str):

        try:
            kms_client = self._get_client()
            parent = f'projects/{project_id}'
            locations = kms_client.projects().locations()
            request = locations.list(name=parent)
            return await GCPFacadeUtils.get_all('locations', request, locations)
        except Exception as e:
            print_exception(f'Failed to retrieve KMS locations: {e}')
            return []

    async def list_key_rings(self, project_id: str):

        try:
            locations = await self.get_locations(project_id)
            key_rings = {}
            for l in locations:
                parent = self.cloud_client.location_path(project_id, l['locationId'])
                key_rings[l['locationId']] = await run_concurrently(
                    lambda: list(self.cloud_client.list_key_rings(parent)))
            return key_rings
        except Exception as e:
            print_exception(f'Failed to retrieve KMS key rings: {e}')
            return {}

    async def list_keys(self, project_id: str, location: str, keyring_name: str):

        try:
            parent = self.cloud_client.key_ring_path(project_id, location, keyring_name)
            kms_client = self._get_client()
            cryptokeys = kms_client.projects().locations().keyRings().cryptoKeys()
            request = cryptokeys.list(parent=parent)
            return await GCPFacadeUtils.get_all('cryptoKeys', request, cryptokeys)
        except Exception as e:
            print_exception(f'Failed to retrieve KMS keys for key ring {keyring_name}: {e}')
            return []

    async def keys_iam_policy(self, project_id: str, location: str, keyring_name: str, key_name: str):

        try:
            parent = self.cloud_client.crypto_key_path(project_id, location, keyring_name, key_name)
            kms_client = self._get_client()
            cryptokeys = kms_client.projects().locations().keyRings().cryptoKeys()
            request = cryptokeys.getIamPolicy(resource=parent)
            return await GCPFacadeUtils.get_all('bindings', request, cryptokeys)
        except Exception as e:
            print_exception(f'Failed to retrieve KMS binding policy for key {key_name}: {e}')
            return []



from azure.mgmt.network import NetworkManagementClient

from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.utils import run_concurrently
from ScoutSuite.utils import get_user_agent


class NetworkFacade:

    def __init__(self, credentials):
        self.credentials = credentials

    def get_client(self, subscription_id: str):
        client = NetworkManagementClient(self.credentials.get_credentials('arm'),
                                       subscription_id=subscription_id)
        client._client.config.add_user_agent(get_user_agent())
        return client

    async def get_network_watchers(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: list(client.network_watchers.list_all())
            )
        except Exception as e:
            print_exception(f'Failed to retrieve network watchers: {e}')
            return []

    async def get_network_security_groups(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: list(client.network_security_groups.list_all())
            )
        except Exception as e:
            print_exception(f'Failed to retrieve network security groups: {e}')
            return []

    async def get_application_security_groups(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: list(client.application_security_groups.list_all())
            )
        except Exception as e:
            print_exception(f'Failed to retrieve application security groups: {e}')
            return []

    async def get_virtual_networks(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: list(client.virtual_networks.list_all())
            )
        except Exception as e:
            print_exception(f'Failed to retrieve virtual networks: {e}')
            return []

    async def get_network_interfaces(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: list(client.network_interfaces.list_all())
            )
        except Exception as e:
            print_exception(f'Failed to retrieve network interfaces: {e}')
            return []



from azure.mgmt.compute import ComputeManagementClient

from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.utils import run_concurrently
from ScoutSuite.utils import get_user_agent


class VirtualMachineFacade:

    def __init__(self, credentials):
        self.credentials = credentials

    def get_client(self, subscription_id: str):
        client = ComputeManagementClient(self.credentials.get_credentials('arm'),
                                       subscription_id=subscription_id)
        client._client.config.add_user_agent(get_user_agent())
        return client

    async def get_instances(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: list(client.virtual_machines.list_all())
            )
        except Exception as e:
            print_exception(f'Failed to retrieve virtual machines: {e}')
            return []

    async def get_instance_extensions(self, subscription_id: str,
                                      instance_name: str,
                                      resource_group: str):
        try:
            client = self.get_client(subscription_id)
            extensions = await run_concurrently(
                lambda: client.virtual_machine_extensions.list(resource_group,
                                                               instance_name)
            )
            return list(extensions.value)
        except Exception as e:
            print_exception(f'Failed to retrieve virtual machine extensions: {e}')
            return []

    async def get_disks(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: list(client.disks.list())
            )
        except Exception as e:
            print_exception(f'Failed to retrieve disks: {e}')
            return []

    async def get_snapshots(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: list(client.snapshots.list())
            )
        except Exception as e:
            print_exception(f'Failed to retrieve snapshots: {e}')
            return []

    async def get_images(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: list(client.images.list())
            )
        except Exception as e:
            print_exception(f'Failed to retrieve images: {e}')
            return []



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.base import AWSCompositeResources
from ScoutSuite.providers.utils import get_non_provider_id
from ScoutSuite.providers.aws.utils import format_arn

from .cluster_parameters import ClusterParameters


class ClusterParameterGroups(AWSCompositeResources):
    _children = [
        (ClusterParameters, 'parameters')
    ]

    def __init__(self, facade: AWSFacade, region: str):
        super().__init__(facade)
        self.region = region
        self.partition = facade.partition
        self.service = 'redshift'
        self.resource_type = 'parametergroup'

    async def fetch_all(self):
        raw_parameter_groups = await self.facade.redshift.get_cluster_parameter_groups(self.region)
        for raw_parameter_group in raw_parameter_groups:
            id, parameter_group = self._parse_parameter_group(raw_parameter_group)
            self[id] = parameter_group

        await self._fetch_children_of_all_resources(
            resources=self,
            scopes={parameter_group_id: {'region': self.region,
                                         'parameter_group_name': parameter_group['name']}
                    for (parameter_group_id, parameter_group) in self.items()}
        )

    def _parse_parameter_group(self, raw_parameter_group):
        parameter_group = {}
        parameter_group['name'] = raw_parameter_group.get('ParameterGroupName')
        parameter_group['id'] = get_non_provider_id(parameter_group['name'])
        parameter_group['arn'] = format_arn(self.partition, self.service, self.region, self.facade.owner_id, raw_parameter_group.get('ParameterGroupName'), self.resource_type)
        parameter_group['family'] = raw_parameter_group.get('ParameterGroupFamily')
        parameter_group['description'] = raw_parameter_group.get('Description')
        parameter_group['is_default'] = self._is_default(raw_parameter_group)
        parameter_group['tags'] = raw_parameter_group.get('Tags')
        return parameter_group['id'], parameter_group

    def _is_default(self, raw_parameter_group):
        if 'Default parameter group for' in raw_parameter_group.get('Description') and \
                'default.' in raw_parameter_group.get('ParameterGroupName'):
            return True
        else:
            return False



from ScoutSuite.providers.aws.resources.base import AWSResources
from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.utils import get_name, format_arn


class Snapshots(AWSResources):
    def __init__(self, facade: AWSFacade, region: str):
        super().__init__(facade)
        self.region = region
        self.partition = facade.partition
        self.service = 'ec2'
        self.resource_type = 'snapshot'

    async def fetch_all(self):
        raw_snapshots = await self.facade.ec2.get_snapshots(self.region)
        for raw_snapshot in raw_snapshots:
            name, resource = self._parse_snapshot(raw_snapshot)
            self[name] = resource

    def _parse_snapshot(self, raw_snapshot):
        snapshot_dict = {}
        snapshot_dict['id'] = raw_snapshot.get('SnapshotId')
        snapshot_dict['name'] = get_name(raw_snapshot, raw_snapshot, 'SnapshotId')
        snapshot_dict['description'] = raw_snapshot.get('Description')
        snapshot_dict['public'] = self._is_public(raw_snapshot)
        snapshot_dict['encrypted'] = raw_snapshot.get('Encrypted')
        snapshot_dict['kms_key_id'] = raw_snapshot.get('KmsKeyId')
        snapshot_dict['owner_id'] = raw_snapshot.get('OwnerId')
        snapshot_dict['progress'] = raw_snapshot.get('Progress')
        snapshot_dict['start_time'] = raw_snapshot.get('StartTime')
        snapshot_dict['state'] = raw_snapshot.get('State')
        snapshot_dict['volume_id'] = raw_snapshot.get('VolumeId')
        snapshot_dict['volume_size'] = raw_snapshot.get('VolumeSize')
        snapshot_dict['create_volume_permissions'] = raw_snapshot.get('CreateVolumePermissions')
        snapshot_dict['arn'] = format_arn(self.partition, self.service, self.region, raw_snapshot.get('OwnerId'), raw_snapshot.get('SnapshotId'), self.resource_type)
        return snapshot_dict['id'], snapshot_dict

    @staticmethod
    def _is_public(snapshot):
        return any([permission.get('Group') == 'all' for permission in snapshot['CreateVolumePermissions']])



import datetime

from azure.mgmt.monitor import MonitorManagementClient
from azure.mgmt.storage import StorageManagementClient

from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.utils import run_concurrently, get_and_set_concurrently
from ScoutSuite.utils import get_user_agent


class StorageAccountsFacade:

    def __init__(self, credentials):
        self.credentials = credentials

    def get_client(self, subscription_id: str):
        client = StorageManagementClient(self.credentials.get_credentials('arm'),
                                       subscription_id=subscription_id)
        client._client.config.add_user_agent(get_user_agent())
        return client

    async def get_storage_accounts(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            storage_accounts = await run_concurrently(
                lambda: list(client.storage_accounts.list())
            )
        except Exception as e:
            print_exception(f'Failed to retrieve storage accounts: {e}')
            return []
        else:
            await get_and_set_concurrently([self._get_and_set_activity_logs], storage_accounts,
                                           subscription_id=subscription_id)
            return storage_accounts

    async def get_blob_containers(self, resource_group_name, storage_account_name, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            containers = await run_concurrently(
                lambda: list(client.blob_containers.list(resource_group_name, storage_account_name))
            )
        except Exception as e:
            print_exception(f'Failed to retrieve blob containers: {e}')
            return []
        else:
            return containers

    async def _get_and_set_activity_logs(self, storage_account, subscription_id: str):
        client = MonitorManagementClient(self.credentials.arm_credentials, subscription_id)
        client._client.config.add_user_agent(get_user_agent())

        # Time format used by Azure API:
        time_format = "%Y-%m-%dT%H:%M:%S.%f"
        # Azure API uses UTC time, we need to use the same to avoid bad requests:
        utc_now = datetime.datetime.utcnow()
        # Activity logs are only archived for a period of 90 days max (requesting a timespan of more than that ends up
        # with a bad request):
        timespan = datetime.timedelta(90)

        logs_filter = " and ".join([
            "eventTimestamp ge {}".format((utc_now - timespan).strftime(time_format)),
            "eventTimestamp le {}".format(utc_now.strftime(time_format)),
            f"resourceId eq {storage_account.id}",
        ])
        try:
            activity_logs = await run_concurrently(
                lambda: list(client.activity_logs.list(filter=logs_filter, select="eventTimestamp, operationName"))
            )
        except Exception as e:
            print_exception(f'Failed to retrieve activity logs: {e}')
            setattr(storage_account, 'activity_logs', [])
        else:
            setattr(storage_account, 'activity_logs', activity_logs)

    # FIXME - not implemented by SDK
    # async def get_queues(self, resource_group_name, storage_account_name, subscription_id: str):
    #     try:
    #         client = QueueServiceClient(account_url="https://{}.queue.core.windows.net".format(storage_account_name),
    #                                      credential=self.credentials.aad_graph_credentials)
    #
    #         # queues = await run_concurrently(
    #         #     lambda: list(client.queues.list(resource_group_name, storage_account_name))
    #         # )
    #     except Exception as e:
    #         print_exception('Failed to retrieve queues: {}'.format(e))
    #         return []
    #     else:
    #         return None
    #         # return queues




from ScoutSuite.core.console import print_exception

def is_throttled(e):
    """
    Determines whether the exception is due to API throttling.

    :param e:                           Exception raised
    :return:                            True if it's a throttling exception else False
    """
    try:
        if 'Quota exceeded' in str(e):
            return True
        else:
            return False
    except Exception as e:
        print_exception(f'Unable to validate exception for throttling: {e}')
        return False



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.base import AWSCompositeResources
from .buckets import Buckets


class S3(AWSCompositeResources):
    _children = [
        (Buckets, 'buckets')
    ]

    def __init__(self, facade: AWSFacade):
        super().__init__(facade)
        self.service = 's3'

    async def fetch_all(self, partition_name='aws', **kwargs):
        # Keep track of regions as S3 is both a global and regional service
        self.facade.s3.regions = kwargs.get('regions')
        self['public_access_block_configuration'] = self.facade.s3.get_s3_public_access_block(self.facade.owner_id)
        await self._fetch_children(self)

    async def finalize(self):
        for bucket_id in self['buckets']:
            if "public_access_block_configuration" in self['buckets'][bucket_id]:
                # The resulting configuration will be the most restrictive
                self['buckets'][bucket_id]["public_access_block_configuration"]["BlockPublicAcls"] = \
                self['buckets'][bucket_id]["public_access_block_configuration"]["BlockPublicAcls"] or \
                self['public_access_block_configuration']["BlockPublicAcls"]
                self['buckets'][bucket_id]["public_access_block_configuration"]["IgnorePublicAcls"] = \
                self['buckets'][bucket_id]["public_access_block_configuration"]["IgnorePublicAcls"] or \
                self['public_access_block_configuration']["IgnorePublicAcls"]
                self['buckets'][bucket_id]["public_access_block_configuration"]["BlockPublicPolicy"] = \
                self['buckets'][bucket_id]["public_access_block_configuration"]["BlockPublicPolicy"] or \
                self['public_access_block_configuration']["BlockPublicPolicy"]
                self['buckets'][bucket_id]["public_access_block_configuration"]["RestrictPublicBuckets"] = \
                self['buckets'][bucket_id]["public_access_block_configuration"]["RestrictPublicBuckets"] or \
                self['public_access_block_configuration']["RestrictPublicBuckets"]
            else:
                # No bucket-level configuration, use account level configuration
                self['buckets'][bucket_id]["public_access_block_configuration"] = self['public_access_block_configuration']



import os

from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.base.provider import BaseProvider
from ScoutSuite.providers.gcp.services import GCPServicesConfig


class GCPProvider(BaseProvider):
    """
    Implements provider for GCP
    """

    def __init__(self,
                 project_id=None, folder_id=None, organization_id=None, all_projects=None,
                 report_dir=None, timestamp=None, services=None, skipped_services=None, result_format='json', **kwargs):
        services = [] if services is None else services
        skipped_services = [] if skipped_services is None else skipped_services

        self.metadata_path = '%s/metadata.json' % os.path.split(
            os.path.abspath(__file__))[0]

        self.provider_code = 'gcp'
        self.provider_name = 'Google Cloud Platform'
        self.environment = 'default'

        self.all_projects = all_projects
        self.project_id = project_id
        self.folder_id = folder_id
        self.organization_id = organization_id

        self.credentials = kwargs['credentials']
        self._set_account_id()

        self.services = GCPServicesConfig(self.credentials, self.credentials.default_project_id,
                                          self.project_id, self.folder_id, self.organization_id, self.all_projects)

        self.result_format = result_format

        super().__init__(report_dir, timestamp,
                         services, skipped_services, result_format)

    def get_report_name(self):
        """
        Returns the name of the report using the provider's configuration
        """
        if self.account_id:
            return f'gcp-{self.account_id}'
        else:
            return 'gcp'

    def _set_account_id(self):
        # All accessible projects
        if self.all_projects:
            # Service Account
            if self.credentials.is_service_account and hasattr(self.credentials, 'service_account_email'):
                self.account_id = self.credentials.service_account_email
            else:
                # TODO use username email (can't find it...)
                self.account_id = 'user-account'
        # Project passed through the CLI
        elif self.project_id:
            self.account_id = self.project_id
        # Folder passed through the CLI
        elif self.folder_id:
            self.account_id = self.folder_id
        # Organization passed through the CLI
        elif self.organization_id:
            self.account_id = self.organization_id
        # Project inferred from default configuration
        elif self.credentials.default_project_id:
            self.account_id = self.credentials.default_project_id
        else:
            self.account_id = 'unknown-project-id'

    def preprocessing(self, ip_ranges=None, ip_ranges_name_key=None):
        """
        Tweak the GCP config to match cross-resources and clean any fetching artifacts

        :param ip_ranges:
        :param ip_ranges_name_key:
        :return: None
        """

        self._match_instances_and_snapshots()
        self._match_networks_and_instances()
        self._match_networks_and_firewalls()
        self._match_subnetworks_and_instances()

        super().preprocessing()

    def _match_instances_and_snapshots(self):
        """
        Compare Compute Engine instances and snapshots to identify instance disks that do not have a snapshot.

        :return:
        """

        try:
            if 'computeengine' in self.service_list:
                for project in self.services['computeengine']['projects'].values():
                    for zone in project['zones'].values():
                        for instance in zone['instances'].values():
                            for instance_disk in instance['disks'].values():
                                instance_disk['snapshots'] = []
                                for disk in project['snapshots'].values():
                                    if disk['status'] == 'READY' and \
                                            disk['source_disk_url'] == instance_disk['source_url']:
                                        instance_disk['snapshots'].append(disk)

                                instance_disk['latest_snapshot'] = max(instance_disk['snapshots'],
                                                                       key=lambda x: x['creation_timestamp']) \
                                    if instance_disk['snapshots'] else None
        except Exception as e:
            print_exception('Unable to match instances and snapshots: {}'.format(e))

    def _match_networks_and_instances(self):
        """
        For each network, math instances in that network

        :return:
        """

        try:
            if 'computeengine' in self.service_list:
                for project in self.services['computeengine']['projects'].values():
                    for network in project['networks'].values():
                        network['instances'] = []
                        for zone in project['zones'].values():
                            # Skip the counts contained in the zones dictionary
                            if zone is int:
                                continue
                            for instance in zone['instances'].values():
                                instance['network_id'] = None
                                for network_interface in instance['network_interfaces']:
                                    if network_interface['network'] == network['network_url']:
                                        network['instances'].append({'instance_id': instance['id'],
                                                                     'instance_zone': instance['zone']})
                                        network_interface['network_id'] = network['id']
        except Exception as e:
            print_exception('Unable to match instances and networks: {}'.format(e))

    def _match_networks_and_firewalls(self):
        """
        For each network, math firewall rules in that network

        :return:
        """

        try:
            if 'computeengine' in self.service_list:
                for project in self.services['computeengine']['projects'].values():
                    for network in project['networks'].values():
                        network['firewalls'] = []
                        for firewall in project['firewalls'].values():
                            firewall['network_id'] = None
                            if firewall['network_url'] == network['network_url']:
                                network['firewalls'].append(firewall['id'])
                                firewall['network_id'] = network['id']
        except Exception as e:
            print_exception('Unable to match firewalls and networks: {}'.format(e))

    def _match_subnetworks_and_instances(self):
        """
        For each subnetwork, math instances in that subnetwork

        :return:
        """

        try:
            if 'computeengine' in self.service_list:
                for project in self.services['computeengine']['projects'].values():
                    for region in project['regions'].values():
                        for subnetwork in region['subnetworks'].values():
                            subnetwork['instances'] = []
                            for zone in project['zones'].values():
                                # Skip the counts contained in the zones dictionary
                                if zone is int:
                                    continue
                                for instance in zone['instances'].values():
                                    instance['subnetwork_id'] = None
                                    for network_interface in instance['network_interfaces']:
                                        if network_interface['subnetwork'] == subnetwork['subnetwork_url']:
                                            subnetwork['instances'].append({'instance_id': instance['id'],
                                                                         'instance_zone': instance['zone']})
                                            network_interface['subnetwork_id'] = subnetwork['id']
                                            network_interface['subnetwork_region'] = subnetwork['region']
        except Exception as e:
            print_exception('Unable to match instances and subnetworks: {}'.format(e))



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.regions import Regions

from .secrets import Secrets


class SecretsManager(Regions):
    _children = [
        (Secrets, 'secrets')
    ]

    def __init__(self, facade: AWSFacade):
        super().__init__('secretsmanager', facade)



from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureResources
from ScoutSuite.providers.utils import get_non_provider_id
from ScoutSuite.providers.azure.utils import get_resource_group_name


class Vaults(AzureResources):

    def __init__(self, facade: AzureFacade, subscription_id: str):
        super().__init__(facade)
        self.subscription_id = subscription_id

    async def fetch_all(self):
        for raw_vault in await self.facade.keyvault.get_key_vaults(self.subscription_id):
            id, vault = self._parse_key_vault(raw_vault)
            self[id] = vault

    def _parse_key_vault(self, raw_vault):
        vault = {}
        vault['id'] = get_non_provider_id(raw_vault.id)
        vault['name'] = raw_vault.name
        vault['type'] = raw_vault.type
        vault['location'] = raw_vault.location
        vault['additional_properties'] = raw_vault.additional_properties
        if raw_vault.tags is not None:
            vault['tags'] = ["{}:{}".format(key, value) for key, value in  raw_vault.tags.items()]
        else:
            vault['tags'] = []
        vault['resource_group_name'] = get_resource_group_name(raw_vault.id)
        vault['properties'] = raw_vault.properties
        vault['public_access_allowed'] = self._is_public_access_allowed(raw_vault)
        return vault['id'], vault

    def _is_public_access_allowed(self, raw_vault):
        return raw_vault.properties.network_acls is None



from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.aws.facade.basefacade import AWSBaseFacade
from ScoutSuite.providers.aws.facade.utils import AWSFacadeUtils
from ScoutSuite.providers.utils import run_concurrently, get_and_set_concurrently
import json



class KMSFacade(AWSBaseFacade):

    async def get_keys(self, region: str):

        try:
            keys = await AWSFacadeUtils.get_all_pages('kms', region, self.session, 'list_keys', 'Keys')
            await get_and_set_concurrently(
                [self._get_and_set_key_policy,
                 self._get_and_set_key_metadata,
                 self._get_and_set_key_aliases],
                keys, region=region)
        except Exception as e:
            print_exception(f'Failed to get KMS keys: {e}')
            keys = []
        finally:
            return keys

    async def _get_and_set_key_policy(self, key: {}, region: str):
        client = AWSFacadeUtils.get_client('kms', self.session, region)
        try:
            response = await run_concurrently(
                lambda: client.get_key_policy(KeyId=key['KeyId'],
                                              PolicyName='default'))
            key['policy'] = json.loads(response.get('Policy'))
        except Exception as e:
            print_exception(f'Failed to get KMS key policy: {e}')

    async def _get_and_set_key_metadata(self, key: {}, region: str):
        client = AWSFacadeUtils.get_client('kms', self.session, region)
        try:
            key['metadata'] = await run_concurrently(lambda: client.describe_key(KeyId=key['KeyId']))
        except Exception as e:
            print_exception(f'Failed to describe KMS key: {e}')

    async def _get_and_set_key_aliases(self, key: {}, region: str):
        client = AWSFacadeUtils.get_client('kms', self.session, region)
        try:
            response = await run_concurrently(
                lambda: client.list_aliases(KeyId=key['KeyId'])
            )
            key['aliases'] = response.get('Aliases')
        except Exception as e:
            print_exception(f'Failed to get KMS aliases: {e}')

    async def get_grants(self, region: str, key_id: str):
        try:
            return await AWSFacadeUtils.get_all_pages('kms', region, self.session, 'list_grants', 'Grants',
                                                      KeyId=key_id)
        except Exception as e:
            print_exception(f'Failed to list KMS Grants: {e}')
            return []

    async def get_key_rotation_status(self, region: str, key_id: str):
        client = AWSFacadeUtils.get_client('kms', self.session, region)
        try:
            return await run_concurrently(
                lambda: client.get_key_rotation_status(KeyId=key_id))
        except Exception as e:
            print_exception(f'Failed to get KMS key rotation: {e}')



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.base import AWSResources


class Snapshots(AWSResources):
    def __init__(self, facade: AWSFacade, region: str, vpc: str):
        super().__init__(facade)
        self.region = region
        self.vpc = vpc

    async def fetch_all(self):
        raw_snapshots = await self.facade.rds.get_snapshots(self.region, self.vpc)
        for raw_snapshot in raw_snapshots:
            name, resource = self._parse_snapshot(raw_snapshot)
            self[name] = resource

    def _parse_snapshot(self, raw_snapshot):
        is_cluster = 'DBClusterIdentifier' in raw_snapshot

        snapshot_id = raw_snapshot.pop('DBClusterSnapshotIdentifier') if is_cluster \
            else raw_snapshot.pop('DBSnapshotIdentifier')

        snapshot = {}
        snapshot['arn'] = raw_snapshot.pop('DBClusterSnapshotArn') if is_cluster else raw_snapshot.pop('DBSnapshotArn')
        snapshot['id'] = snapshot_id,
        snapshot['name'] = snapshot_id,
        snapshot['vpc_id'] = raw_snapshot['VpcId']
        snapshot['attributes'] = raw_snapshot['Attributes']
        snapshot['is_cluster'] = is_cluster

        attributes = [
            'DBInstanceIdentifier',
            'SnapshotCreateTime',
            'Encrypted',
            'OptionGroupName'
        ]
        for attribute in attributes:
            snapshot[attribute] = raw_snapshot[attribute] if attribute in raw_snapshot else None

        if snapshot['is_cluster']:  # Map some fields to do more generic and simple rules
            snapshot['DBClusterIdentifier'] = raw_snapshot['DBClusterIdentifier']
            snapshot['Encrypted'] = raw_snapshot['StorageEncrypted']

        return snapshot_id, snapshot



import re
from ScoutSuite.core.console import print_exception

ec2_classic = "EC2-Classic"


def get_caller_identity(session):
    sts_client = session.client("sts")
    identity = sts_client.get_caller_identity()
    return identity


def get_aws_account_id(session):
    caller_identity = get_caller_identity(session)
    account_id = caller_identity["Arn"].split(":")[4]
    return account_id


def get_partition_name(session):
    caller_identity = get_caller_identity(session)
    partition_name = caller_identity["Arn"].split(":")[1]
    return partition_name


def is_throttled(e):
    """
    Determines whether the exception is due to API throttling.

    :param e:                           Exception raised
    :return:                            True if it's a throttling exception else False
    """
    try:
        return (
            hasattr(e, "response")
            and e.response
            and "Error" in e.response
            and e.response["Error"]["Code"]
            in ["Throttling", "RequestLimitExceeded", "ThrottlingException"]
        )
    except Exception as e:
        print_exception(f'Unable to validate exception for throttling: {e}')
        return False


def get_keys(src, dst, keys):
    """
    Copies the value of keys from source object to dest object

    :param src:                         Source object
    :param dst:                         Destination object
    :param keys:                        Keys
    :return:
    """
    for key in keys:
        dst[key] = src[key] if key in src else None


def get_name(src, dst, default_attribute):
    """

    :param src:                         Source object
    :param dst:                         Destination object
    :param default_attribute:           Default attribute

    :return:
    """
    name_found = False
    if "Tags" in src:
        for tag in src["Tags"]:
            if tag["Key"] == "Name" and tag["Value"] != "":
                dst["name"] = tag["Value"]
                name_found = True
    if not name_found:
        dst["name"] = src[default_attribute]
    return dst["name"]


def no_camel(name):
    """
    Converts CamelCase to camel_case

    :param name:                        Name string to convert
    :return:
    """
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


def snake_keys(d):
    """
    Converts a dictionary with CamelCase keys to camel_case

    :param name:                        d Dictionary to iterate over
    :return:
    """

    new_table = {}
    if isinstance(d, dict):
        for k in d.keys():
            new_key = no_camel(k)
            if isinstance(d[k], dict):
                new_table[new_key] = snake_keys(d[k])
            elif isinstance(d[k], list):
                new_ary = []
                for v in d[k]:
                    if isinstance(v, dict):
                        new_ary.append(snake_keys(v))
                    else:
                        new_ary.append(v)
                new_table[new_key] = new_ary
            else:
                new_table[new_key] = d[k]
    return new_table

def format_arn(partition, service, region, account_id, resource_id, resource_type=None):
    """
    Formats a resource ARN based on the parameters

    :param partition:                   The partition where the resource is located
    :param service:                     The service namespace that identified the AWS product
    :param region:                      The corresponding region
    :param account_id:                  The ID of the AWS account that owns the resource
    :param resource_id:                 The resource identified
    :param resource_type:               (Optional) The resource type
    :return:                            Resource ARN
    """

    try:
        # If a resource type is specified
        if resource_type is not None:
            arn = f"arn:{partition}:{service}:{region}:{account_id}:{resource_type}/{resource_id}"
        else:
            arn = f"arn:{partition}:{service}:{region}:{account_id}:{resource_id}"
    except Exception as e:
        print_exception(f'Failed to parse a resource ARN: {e}')
        return None
    return arn



from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureResources
from ScoutSuite.providers.utils import get_non_provider_id


class ComplianceResults(AzureResources):

    def __init__(self, facade: AzureFacade, subscription_id: str):
        super().__init__(facade)
        self.subscription_id = subscription_id

    async def fetch_all(self):
        for raw_compliance_result in await self.facade.securitycenter.get_compliance_results(self.subscription_id):
            id, compliance_result = self._parse_compliance_result(raw_compliance_result)
            self[id] = compliance_result

    def _parse_compliance_result(self, raw_compliance_result):
        compliance_result_dict = {}
        compliance_result_dict['id'] = get_non_provider_id(raw_compliance_result.id)
        compliance_result_dict['name'] = raw_compliance_result.name
        compliance_result_dict['type'] = raw_compliance_result.type
        compliance_result_dict['resource_status'] = raw_compliance_result.resource_status
        compliance_result_dict['additional_properties'] = raw_compliance_result.additional_properties
        return compliance_result_dict['id'], compliance_result_dict



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.base import AWSResources
from ScoutSuite.providers.utils import get_non_provider_id
from ScoutSuite.providers.aws.utils import format_arn


class MetricFilters(AWSResources):
    def __init__(self, facade: AWSFacade, region: str):
        super(MetricFilters, self).__init__(facade)
        self.region = region
        self.partition = facade.partition
        self.service = 'cloudwatch'
        self.resource_type = 'metric-filter'

    async def fetch_all(self):
        for raw_metric_filter in await self.facade.cloudwatch.get_metric_filters(self.region):
            name, resource = self._parse_metric_filter(raw_metric_filter)
            self[name] = resource

    def _parse_metric_filter(self, raw_metric_filter):
        metric_filter_dict = {}
        metric_filter_dict['id'] = get_non_provider_id('{}{}'.format(raw_metric_filter.get('filterName'),
                                                                     raw_metric_filter.get('creationTime')))
        metric_filter_dict['name'] = raw_metric_filter.get('filterName')
        metric_filter_dict['creation_time'] = raw_metric_filter.get('creationTime')
        metric_filter_dict['pattern'] = raw_metric_filter.get('filterPattern')
        metric_filter_dict['metric_transformations'] = raw_metric_filter.get('metricTransformations')
        metric_filter_dict['log_group_name'] = raw_metric_filter.get('logGroupName')
        metric_filter_dict['arn'] = format_arn(self.partition, self.service, self.region, '', raw_metric_filter.get('filterName'), self.resource_type)
        return metric_filter_dict['id'], metric_filter_dict





from ScoutSuite.providers.base.resources.base import Resources
from ScoutSuite.providers.gcp.facade.base import GCPFacade


class ManagedZones(Resources):
    def __init__(self, facade: GCPFacade, project_id: str):
        super().__init__(facade)
        self.project_id = project_id

    async def fetch_all(self):
        raw_zones = await self.facade.dns.get_zones(self.project_id)
        for raw_zone in raw_zones['managedZones']:
            zone_id, zone = self._parse_zone(raw_zone)
            self[zone_id] = zone

    def _parse_zone(self, raw_zone):
        zone_dict = {}
        zone_dict['id'] = raw_zone['id']
        zone_dict['name'] = raw_zone['name']
        zone_dict['description'] = self._get_description(raw_zone)
        zone_dict['dns_name'] = raw_zone['dnsName']
        zone_dict['name_servers'] = raw_zone.get('nameServers', None)
        zone_dict['visibility'] = raw_zone['visibility']
        zone_dict['creation_timestamp'] = raw_zone['creationTime']

        dnssec_config = raw_zone.get('dnssecConfig',None)
        if dnssec_config:
            zone_dict['dnssec_enabled'] = True if dnssec_config['state'] == 'on' else False
            zone_dict['dnssec_keys'] = self._get_keys(dnssec_config,zone_dict)
        else:
            zone_dict['dnssec_enabled'] = False
            zone_dict['dnssec_keys'] = None
            zone_dict['key_signing_algorithm'] = None
            zone_dict['zone_signing_algorithm']=None
        return zone_dict['id'], zone_dict

    def _get_description(self, raw_zone):
        description = raw_zone.get('description')
        return description if description else 'N/A'

    def _get_keys(self, dnssec_config,zone_dict):
        raw_keys = dnssec_config.get('defaultKeySpecs', None)
        if not raw_keys:
            return None
        key_dict = {}
        for raw_key in raw_keys:
            key_dict[raw_key['keyType']]={
                'key_type': raw_key['keyType'],
                'key_algorithm': raw_key['algorithm'],
                'length': raw_key['keyLength'],
            }
            if raw_key['keyType'] == 'keySigning':
                zone_dict['key_signing_algorithm'] = raw_key['algorithm']
            elif raw_key['keyType'] == 'zoneSigning':
                zone_dict['zone_signing_algorithm'] = raw_key['algorithm']



        return key_dict



from ScoutSuite.providers.azure.resources.base import AzureCompositeResources
from ScoutSuite.core.console import print_exception

from .users import Users
from .groups import Groups
from .serviceprincipals import ServicePrincipals
from .applications import Applications


class AAD(AzureCompositeResources):
    _children = [
        (Users, 'users'),
        (Groups, 'groups'),
        (ServicePrincipals, 'service_principals'),
        (Applications, 'applications'),
    ]

    async def fetch_all(self):
        await self._fetch_children(resource_parent=self)

    async def fetch_additional_users(self, user_list):
        """
        Special method to fetch additional users
        """
        try:
            # fetch the users
            additional_users = Users(self.facade)
            await additional_users.fetch_additional_users(user_list)
            # add them to the resource and update count
            self['users'].update(additional_users)
            self['users_count'] = len(self['users'].values())
        except Exception as e:
            print_exception('Unable to fetch additional users: {}'.format(e))
        finally:
            # re-run the finalize method
            await self.finalize()

    async def finalize(self):
        self.assign_group_memberships()

    def assign_group_memberships(self):
        """
        Assigns members to groups
        """
        try:
            for group in self['groups']:
                for user in self['users']:
                    if group in self['users'][user]['groups']:
                        self['groups'][group]['users'].append(user)
        except Exception as e:
            print_exception('Unable to assign group memberships: {}'.format(e))



from ScoutSuite.providers.base.resources.base import Resources
from ScoutSuite.providers.gcp.facade.base import GCPFacade


class LoggingMetrics(Resources):
    def __init__(self, facade: GCPFacade, project_id: str):
        super().__init__(facade)
        self.project_id = project_id

    async def fetch_all(self):
        raw_metrics = await self.facade.stackdriverlogging.get_metrics(self.project_id)
        metric = self._parse_metric(raw_metrics)
        self[self.project_id] = metric

    def _parse_metric(self, raw_metrics):
        metric_dict = {}
        metric_dict['project_ownership_assignments'] =\
            self._specific_filter_present(raw_metrics, '(protoPayload.serviceName="cloudresourcemanager.googleapis'
                                                       '.com") AND (ProjectOwnership OR projectOwnerInvitee) OR ('
                                                       'protoPayload.serviceData.policyDelta.bindingDeltas.action'
                                                       '="REMOVE" AND '
                                                       "protoPayload.serviceData.policyDelta.bindingDeltas.role"
                                                       '="roles/owner") OR ('
                                                       'protoPayload.serviceData.policyDelta.bindingDeltas.action'
                                                       '="ADD" AND '
                                                       'protoPayload.serviceData.policyDelta.bindingDeltas.role'
                                                       '="roles/owner")')
        metric_dict['audit_config_change'] = \
            self._specific_filter_present(raw_metrics, 'protoPayload.methodName="SetIamPolicy" AND '
                                                       'protoPayload.serviceData.policyDelta.auditConfigDeltas:*')
        metric_dict['custom_role_change'] = \
            self._specific_filter_present(raw_metrics, 'resource.type="iam_role" AND protoPayload.methodName =  '
                                                       '"google.iam.admin.v1.CreateRole" OR '
                                                       'protoPayload.methodName="google.iam.admin.v1.DeleteRole" OR '
                                                       'protoPayload.methodName="google.iam.admin.v1.UpdateRole"')
        metric_dict['vpc_network_firewall_rule_change'] = \
            self._specific_filter_present(raw_metrics, 'resource.type="gce_firewall_rule" AND '
                                                       'jsonPayload.event_subtype="compute.firewalls.patch" OR '
                                                       'jsonPayload.event_subtype="compute.firewalls.insert"')
        metric_dict['vpc_network_route_change'] = \
            self._specific_filter_present(raw_metrics, 'resource.type="gce_route" AND '
                                                       'jsonPayload.event_subtype="compute.routes.delete" OR '
                                                       'jsonPayload.event_subtype="compute.routes.insert"')
        metric_dict['vpc_network_change'] = \
            self._specific_filter_present(raw_metrics, 'resource.type=gce_network AND '
                                                       'jsonPayload.event_subtype="compute.networks.insert" OR '
                                                       'jsonPayload.event_subtype="compute.networks.patch" OR '
                                                       'jsonPayload.event_subtype="compute.networks.delete"  OR '
                                                       'jsonPayload.event_subtype="compute.networks.removePeering" OR '
                                                       'jsonPayload.event_subtype="compute.networks.addPeering"')
        metric_dict['cloud_storage_iam_permission_change'] = \
            self._specific_filter_present(raw_metrics, 'resource.type=gcs_bucket AND '
                                                       'protoPayload.methodName="storage.setIamPermissions"')
        metric_dict['sql_instance_conf_change'] = \
            self._specific_filter_present(raw_metrics, 'protoPayload.methodName="cloudsql.instances.update"')

        return metric_dict

    def _specific_filter_present(self, raw_metrics, filter_value: str):
        for metric in raw_metrics:
            if metric.filter_ == filter_value:
                return True
        return False
